﻿#undef debugTestHandlers
#define debugTestHandlers

using CySmart.Common.Base.Compatibility.Net20;
using CySmart.DongleCommunicator.API;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO.Ports;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;


namespace Host
{
    public static class BleHandlers
    {

        public static void SetupBleMgrHandlers()
        {
            string sm;

#if (newFunc)
            // Reports the status of ICyBleMgr.SetDeviceIoCapabilities()
            Program.mf.bleMgrCb.SetDeviceIoCapabilitiesHandler = (status) =>
            {
            };

            // Reports the IO capabilities of the local device
            Program.mf.bleMgrCb.GetDeviceIoCapabilitiesHandler = (ioCapabilities, status) =>
            {
            };

            // Reports the RSSI and the status of ICyBleMgr.GetRSSI()
            Program.mf.bleMgrCb.GetRssiHandler = (rssi, status) =>
            {
            };
#endif

            // Reports the result and status of ICyBleMgr.Connect()
            Program.mf.bleMgrCb.ConnectedHandler = (result, status) =>
            {
                CyApiErr err;

                if (status == CyStatus.BLE_STATUS_OK)
                {
                    Program.mf.connectionTimer.Stop();
                    sm = string.Format("Pairing with {0:X12}", Program.mf.device.PeerDeviceAddress.Address);
                    Utilities.SetText(Program.mf.statusMessage, sm);

                    Program.mf.connectedDevice = result.Device;
                    err = Program.mf.connectedDevice.RegisterCallback(Program.mf.devCb);
                    if (err.IsNotOK)
                    {
#if (debugTestHandlers)
                        // The following is debug-only nonsense
                        Utilities.ForceException(8);
#endif
                    }
                    else
                    {
                        CyPairSettings settings = new CyPairSettings(CySecurityLevel.AUTHENTICATED_PAIRING_WITH_ENCRYPTION,
//                        CyPairSettings settings = new CyPairSettings(CySecurityLevel.AUTHENTICATED_LE_SECURE_CONNECTION_PAIRING_WITH_ENCRYPTION,
//                        CyPairSettings settings = new CyPairSettings(CySecurityLevel.UNAUTHENTICATED_PAIRING_WITH_ENCRYPTION,
                                                                     true,
                                                                     16,
                                                                     CyPairingProperties.USE_MITM);
//                        CyPairingProperties.NONE); // .MITM??
                        err = Program.mf.connectedDevice.Pair(settings);
                        if (err.IsNotOK)
                        {
#if (debugTestHandlers)
                            // The following is debug-only nonsense
                            Utilities.ForceException(9);
#endif
                        }
                        Utilities.SetState(Constants.hostStates.HOST_STATE_PAIRING);
                    }
                }
                else
                {
#if (debugTestHandlers)
                    // The following is debug-only nonsense
                    Utilities.ForceException(3);
#endif
                    sm = string.Format("Failed to connect to {0:X12} - status: {1}",
                                       Program.mf.device.PeerDeviceAddress.Address, status.ToString());
                    Utilities.SetText(Program.mf.statusMessage, sm);
                    Utilities.SetState(Constants.hostStates.HOST_STATE_READY);
                    Program.mf.connectionTimer.Stop();
                }
            };

            // Reports the status of ICyBleMgr.CancelConnection()
            Program.mf.bleMgrCb.CancelledHandler = (status) =>
            {
                if (status == CyStatus.BLE_STATUS_OK)
                {
                    if (Program.mf.connectionTimeLeft <= 0)
                    {
                        sm = string.Format("Connection to {0:X12} timed out", Program.mf.device.PeerDeviceAddress.Address);
                    }
                    else
                    {
                        sm = string.Format("Canceled connection to {0:X12}", Program.mf.device.PeerDeviceAddress.Address);
                    }
                    Utilities.SetText(Program.mf.statusMessage, sm);
                    Utilities.SetState(Constants.hostStates.HOST_STATE_READY);
                }
                else
                {
#if (debugTestHandlers)
                    // The following is debug-only nonsense
                    Utilities.ForceException(4);
#endif
                    sm = string.Format("Connection to {0:X12} canceled: status = {1}",
                                       Program.mf.device.PeerDeviceAddress.Address, status.ToString());
                    Utilities.SetText(Program.mf.statusMessage, sm);
                }
                Program.mf.connectionTimer.Stop();
            };

            // Reports the status of ICyBleMgr.Disconnect()
            Program.mf.bleMgrCb.DisconnectedHandler = (deviceAddress, status) =>
            {
                if (status == CyStatus.BLE_STATUS_OK)
                {
                    if (!deviceAddress.Equals(Program.mf.device.PeerDeviceAddress))
                    {
#if (debugTestHandlers)
                        // The following is debug-only nonsense
                        Utilities.ForceException(23);
#endif
                    }
                    sm = string.Format("Disconnected from {0:X12}", Program.mf.device.PeerDeviceAddress.Address);
                    Utilities.SetText(Program.mf.statusMessage, sm);
                    Utilities.SetState(Constants.hostStates.HOST_STATE_READY);
                }
                else
                {
#if (debugTestHandlers)
                    // The following is debug-only nonsense
                    Utilities.ForceException(7);
#endif
                    sm = string.Format("Failed to disconnect from {0:X12} - status: {1}",
                                       Program.mf.device.PeerDeviceAddress.Address, status.ToString());
                    Utilities.SetText(Program.mf.statusMessage, sm);
                }
            };

#if (newFunc)
            // Reports the registered PSM
            Program.mf.bleMgrCb.RegisterPsmHandler = (psm, status) =>
            {
            };

            // Reports the status of ICyBleMgr.UnregisterPsm()
            Program.mf.bleMgrCb.UnregisterPsmHandler = (status) =>
            {
            };

            // Reports the status of ICyBleMgr.SetHostChannelClassification()
            Program.mf.bleMgrCb.SetHostChannelClassificationHandler = (status) =>
            {
            };

            // Reports the status of ICyBleMgr.SetTxPower()
            Program.mf.bleMgrCb.SetTxPowerHandler = (status) =>
            {
            };

            // Reports the channel transmission power
            Program.mf.bleMgrCb.GetTxPowerHandler = (result, status) =>
            {
            };

            // Reports the default data length
            Program.mf.bleMgrCb.GetDefaultDataLengthHandler = (result, status) =>
            {
            };

            // Reports the status of ICyBleMgr.SetSuggestedDataLength()
            Program.mf.bleMgrCb.SetSuggestedDataLengthHandler = (status) =>
            {
            };

            // Reports the result of the octet-to-time conversion
            Program.mf.bleMgrCb.ConvertDataLengthOctetToTimeHandler = (result, status) =>
            {
            };

            // Reports the status of ICyBleMgr.SetResolvableAddressTimeout()
            Program.mf.bleMgrCb.SetResolvableAddressTimeoutHandler = (status) =>
            {
            };

            // Reports the status of ICyBleMgr.SetAddressResolutionControl()
            Program.mf.bleMgrCb.SetAddressResolutionControlHandler = (status) =>
            {
            };
#endif
        }

        public static void SetupScanHandlers()
        {
            /// <summary>
            /// Setup the scan result handler to log the discovered BLE devices
            /// </summary>
            Program.mf.scanCb.ScanResultHandler = (result) =>
            {
                if (result != null)
                {
                    foreach (var item in result.ScanRecords)
                    {
                        Utilities.AddDevice(Program.mf.scanResults, item);
                    }
                }
            };

            // scanCb.ScanStatusChangedHandler is not defined here because it is
            //  setup dynamically via StartScan() and StopScan()

            Program.mf.scanCb.StartScanErrorHandler = (status) =>
            {
                string sm = string.Format("Scan start error: {0}", status);
                Utilities.SetText(Program.mf.statusMessage, sm);
                Utilities.SetState(Constants.hostStates.HOST_STATE_READY);
            };

            Program.mf.scanCb.StopScanErrorHandler = (status) =>
            {
                string sm = string.Format("Scan stop error: {0}", status);
                Utilities.SetText(Program.mf.statusMessage, sm);
                Utilities.SetState(Constants.hostStates.HOST_STATE_SCANNING);
            };
        }

        public static void SetupDeviceListHandlers()
        {
#if (newFunc)
            // Reports the device added to the whitelist
            Program.mf.devListCb.AddDeviceToWhitelistHandler = (device, status) =>
            {
            };

            // Reports the status of ICyBleDeviceList.RemoveDeviceFromWhitelist()
            Program.mf.devListCb.RemoveDeviceFromWhitelistHandler = (status) =>
            {
            };

            // Reports the status of ICyBleDeviceList.ClearWhitelist()
            Program.mf.devListCb.ClearWhitelistHandler = (status) =>
            {
            };

            // Reports the whitelist devices
            Program.mf.devListCb.GetWhitelistDevicesHandler = (whitelist, status) =>
            {
            };

            // Reports the bond list devices
            Program.mf.devListCb.GetBondListDevicesHandler = (bondList, status) =>
            {
            };

            // Reports the device added to the resolving list
            Program.mf.devListCb.AddDeviceToResolvingListHandler = (device, status) =>
            {
            };

            // Reports the status of ICyBleDeviceList.RemoveDeviceFromResolvinglist()
            Program.mf.devListCb.RemoveDeviceFromResolvingListHandler = (status) =>
            {
            };

            // Reports the status of ICyBleDeviceList.ClearResolvinglist()
            Program.mf.devListCb.ClearResolvingListHandler = (status) =>
            {
            };

            // Reports the resolving list devices
            Program.mf.devListCb.GetResolvingListDevicesHandler = (resolvingList, status) =>
            {
            };
#endif
        }

        public static void SetupSecurityMgrHandlers()
        {
#if (newFunc)
            // Reports the local authentication keys
            Program.mf.secMgrCb.GetAuthenticationKeysHandler = (result, status) =>
            {
            };

            // Reports the generated authentication keys
            Program.mf.secMgrCb.GenerateAuthenticationKeysHandler = (result, status) =>
            {
            };

            // Reports the status of ICyBleSecurityMgr.SetAuthenticationKeys()
            Program.mf.secMgrCb.SetAuthenticationKeysHandler = (status) =>
            {
            };

            // Reports the peer device authentication keys
            Program.mf.secMgrCb.GetPeerDeviceAuthenticationKeysHandler = (result, status) =>
            {
            };

            // Reports the status of ICyBleSecurityMgr.ResolvePeerDeviceAddress()
            Program.mf.secMgrCb.ResolvePeerDeviceAddressHandler = (result, status) =>
            {
            };

            // Reports the generated secure connection OOB data
            Program.mf.secMgrCb.GenerateSecureConnectionOobDataHandler = (result, status) =>
            {
            };

            // Reports the status of ICyBleSecurityMgr.GenerateLocalECDHKey()
            Program.mf.secMgrCb.GenerateLocalECDHKeyHandler = (status) =>
            {
            };
#endif
        }

        public static void SetupDeviceAddressMgrHandlers()
        {
#if (newFunc)
            // Reports the generated Bluetooth device address
            Program.mf.devAddrMgrCb.BdAddressGeneratedHandler = (result, status) =>
            {
            };

            // Reports the Bluetooth device address of the local device (dongle)
            Program.mf.devAddrMgrCb.GetBdAddressHandler = (address, status) =>
            {
            };

            // Reports the status of ICyBleDeviceAddressMgr.SetBdAddress()
            Program.mf.devAddrMgrCb.SetBdAddressHandler = (status) =>
            {
            };

            // Reports the status of ICyBleDeviceAddressMgr.SetIdAddress()
            Program.mf.devAddrMgrCb.SetIdAddressHandler = (status) =>
            {
            };

            // Reports the resolvable address of the peer device
            Program.mf.devAddrMgrCb.GetPeerResolvableAddressHandler = (result, status) =>
            {
            };

            // Reports the resolvable address of the local device
            Program.mf.devAddrMgrCb.GetLocalResolvableAddressHandler = (result, status) =>
            {
            };
#endif
        }

        public static void SetupDeviceHandlers()
        {
            // Connection parameter update request received from remote device
            Program.mf.devCb.UpdateConnectionParameterRequestReceivedHandler = (request) =>
            {
                CyApiErr err;

                err = Program.mf.connectedDevice.SendConnectionParametersResponse(CyConnectionParametersResponse.ACCEPT);
                if (err.IsNotOK)
                {
#if (debugTestHandlers)
                    // The following is debug-only nonsense
                    Utilities.ForceException(0x00F7);
#endif
                }
            };

#if (newFunc)
            // Reports the status of ICyBleDevice.UpdateConnectionParameter()
            Program.mf.devCb.UpdateConnectionParameterHandler = (status) =>
            {
            };
#endif

            // Reports the updated connection parameter values
            Program.mf.devCb.ConnectionParameterChangedHandler = (connectionParameters) =>
            {
                Program.mf.connectionInterval = connectionParameters.ConnectionInterval;
                Program.mf.slaveLatency = connectionParameters.SlaveLatency;
                Program.mf.supervisionTimeout = connectionParameters.SupervisionTimeout;
            };

#if (newFunc)
            // Reports the status of ICyBleDevice.SetDataLength()
            Program.mf.devCb.SetDataLengthHandler = (status) =>
            {
            };

            // Reports the updated data length for the connection
            Program.mf.devCb.DataLengthChangedHandler = (dataLength) =>
            {
            };

            // Reports the pairing request received from the remote device
            Program.mf.devCb.PairRequestReceivedHandler = (settings) =>
            {
            };

            // Reports the negotiated pairing settings
            Program.mf.devCb.PairingSettingsNegotiatedHandler = (settings) =>
            {
            };

            // Passkey entry request
            Program.mf.devCb.PasskeyEntryRequestHandler = (response) =>
            {
            };

            // Passkey display request
            Program.mf.devCb.PasskeyDisplayRequestHandler = (info) =>
            {
            };

            // Secure connection numeric comparison request
            Program.mf.devCb.NumericComparisonRequestHandler = (response) =>
            {
            };

            // Reports the status of ICyBleDevice.SendKeyPressNotification()
            Program.mf.devCb.SendKeyPressNotificationHandler = (status) =>
            {
            };
#endif

            // Reports the status of ICyBleDevice.Pair()
            Program.mf.devCb.PairingCompletedHandler = (status) =>
            {
                if (status == CyStatus.BLE_STATUS_OK)
                {
                    CyApiErr err = Program.mf.gattClient.RegisterCallback(Program.mf.gattCb);
                    if (err.IsNotOK)
                    {
#if (debugTestHandlers)
                        // The following is debug-only nonsense
                        Utilities.ForceException(11);
#endif
                    }
                    else
                    {
                        err = Program.mf.l2capMgr.RegisterL2CapCallback(Program.mf.l2capMgrCb);
                        if (err.IsNotOK)
                        {
#if (debugTestHandlers)
                            // The following is debug-only nonsense
                            Utilities.ForceException(10);
#endif
                        }
                        else
                        {
                            err = Program.mf.gattClient.DiscoverAllServices();
                            if (err.IsNotOK)
                            {
#if (debugTestHandlers)
                                // The following is debug-only nonsense
                                Utilities.ForceException(17);
#endif
                            }
                            Utilities.SetState(Constants.hostStates.HOST_STATE_DISCOVERING_GATT);
                        }
                    }
                }
                else
                {
#if (debugTestHandlers)
                    // The following is debug-only nonsense
                    Utilities.ForceException(40);
#endif
                    string sm = string.Format("Failed to pair with {0:X12} - status: {1}",
                                              Program.mf.device.PeerDeviceAddress.Address, status.ToString());
                    Utilities.SetText(Program.mf.statusMessage, sm);
                    Utilities.SetState(Constants.hostStates.HOST_STATE_READY);
                }
            };

#if (newFunc)
            // Reports the status of ICyBleDevice.SetOob()
            Program.mf.devCb.SetOobHandler = (status) =>
            {
            };
#endif
        }

        public static void SetupGattHandlers()
        {
#if (newFunc)
            // Reports the negotiated GATT MTU size
            Program.mf.gattCb.GattMtuExchangedHandler = (result, status) =>
            {
            };
#endif

            // Reports all the discovered services
            Program.mf.gattCb.ServiceDiscoveredHandler = (result, status) =>
            {
                string sm;

                if (status == CyStatus.BLE_STATUS_OK)
                {
                    if (result != null)
                    {
                        sm = string.Format("Connected to {0:X12}",
                                           Program.mf.device.PeerDeviceAddress.Address);
                        Utilities.SetText(Program.mf.statusMessage, sm);
                        Utilities.SetState(Constants.hostStates.HOST_STATE_DEVICE_CONNECTED);
                        Program.mf.gattData = result;
                        Globals.uapiHidData.inputs = new List<Globals.hidReportChar>();     // TODO - Disconnection => releasing this memory
                        Globals.uapiHidData.outputs = new List<Globals.hidReportChar>();    // TODO - Disconnection => releasing this memory
                        Gatt.gattDialogBox();
                    }
                }
                else
                {
#if (debugTestHandlers)
                    // The following is debug-only nonsense
                    Utilities.ForceException(20);
#endif
                    sm = string.Format("GATT service discovery error: status = {0}",
                                       status.ToString());
                    Utilities.SetText(Program.mf.statusMessage, sm);
                }
            };

            // Reports the received characteristic value notifications and indications
            Program.mf.gattCb.CharacteristicChangedHandler = (info) =>
            {
                if ((Globals.hidState == Constants.hidStates.HID_STATE_COMPLETE) && (info.Handle == Globals.uapiHidData.inputs[Globals.uapiHidData.inputIdx].dataHdl))
                {
                    Uapi.uapiMessages.AppendText(Utilities.ByteArrayToHexString(info.Value) + "\n", Color.Red);
                }
            };

            // Reports the characteristic value that was read
            Program.mf.gattCb.CharacteristicReadHandler = (result, status) =>
            {
                CyGattReadInfo rdInfo;
                CyGattWriteInfo wrInfo;
                CyApiErr err;

                if (status == CyStatus.BLE_STATUS_OK)
                {
                    if (Globals.hidState == Constants.hidStates.HID_STATE_CONFIRMING_INPUT)
                    {
                        // Is this a valid input report descriptor?
                        if ( (result.Handle == Globals.uapiHidData.inputs[Globals.uapiHidData.mappingHdlIdx].mappingHdl) &&
                             (result.Offset == 0) && (result.Value.Length == 2) &&
                             ((Constants.hidReportTypes)result.Value[1] == Constants.hidReportTypes.HID_REPORT_TYPE_INPUT) )
                        {
                            // Valid, so save its report ID
                            Globals.uapiHidData.inputs[Globals.uapiHidData.mappingHdlIdx].ID = (ushort)result.Value[0];
                            // Is this the UAPI input report descriptor?
                            if ((Constants.hidReportIDs)result.Value[0] == Constants.hidReportIDs.HID_REPORT_ID_UAPI_IN)
                            {
                                // Save the descriptor index for future access
                                Globals.uapiHidData.inputIdx = (short)Globals.uapiHidData.mappingHdlIdx;
                            }
                            if (Globals.uapiHidData.mappingHdlIdx < Globals.uapiHidData.inputs.Count - 1)
                            {
                                Globals.uapiHidData.mappingHdlIdx++;
                                rdInfo = new CyGattReadInfo(Globals.uapiHidData.inputs[Globals.uapiHidData.mappingHdlIdx].mappingHdl);
                                err = Program.mf.gattClient.ReadCharacteristic(rdInfo);
                                if (err.IsNotOK)
                                {
#if (debugTestHandlers)
                                    // The following is debug-only nonsense
                                    Utilities.ForceException(30);
#endif
                                }
                            }
                            else
                            {
                                Globals.hidState = Constants.hidStates.HID_STATE_CONFIRMING_OUTPUT;
                                Uapi.uapiMessages.AppendText("Confirming Output\n", Color.Blue);
                                Globals.uapiHidData.mappingHdlIdx = 0;
                                rdInfo = new CyGattReadInfo(Globals.uapiHidData.outputs[0].mappingHdl);
                                err = Program.mf.gattClient.ReadCharacteristic(rdInfo);
                                if (err.IsNotOK)
                                {
#if (debugTestHandlers)
                                    // The following is debug-only nonsense
                                    Utilities.ForceException(31);
#endif
                                }
                            }
                            return;
                        }
                    }
                    else if (Globals.hidState == Constants.hidStates.HID_STATE_CONFIRMING_OUTPUT)
                    {
                        // Is this a valid output report descriptor?
                        if ( (result.Handle == Globals.uapiHidData.outputs[Globals.uapiHidData.mappingHdlIdx].mappingHdl) &&
                             (result.Offset == 0) && (result.Value.Length == 2) &&
                             ((Constants.hidReportTypes)result.Value[1] == Constants.hidReportTypes.HID_REPORT_TYPE_OUTPUT) )
                        {
                            // Valid, so save its report ID
                            Globals.uapiHidData.outputs[Globals.uapiHidData.mappingHdlIdx].ID = (ushort)result.Value[0];
                            // Is this the UAPI output report descriptor?
                            if ((Constants.hidReportIDs)result.Value[0] == Constants.hidReportIDs.HID_REPORT_ID_UAPI_OUT)
                            {
                                // Save the descriptor index for future access
                                Globals.uapiHidData.outputIdx = (short)Globals.uapiHidData.mappingHdlIdx;
                            }
                            if (Globals.uapiHidData.mappingHdlIdx < Globals.uapiHidData.outputs.Count - 1)
                            {
                                Globals.uapiHidData.mappingHdlIdx++;
                                rdInfo = new CyGattReadInfo(Globals.uapiHidData.outputs[Globals.uapiHidData.mappingHdlIdx].mappingHdl);
                                err = Program.mf.gattClient.ReadCharacteristic(rdInfo);
                                if (err.IsNotOK)
                                {
#if (debugTestHandlers)
                                    // The following is debug-only nonsense
                                    Utilities.ForceException(32);
#endif
                                }
                                return;
                            }
                            else if ( (Globals.uapiHidData.inputIdx != -1) && (Globals.uapiHidData.outputIdx != -1) )
                            {
                                Globals.hidState = Constants.hidStates.HID_STATE_ENABLING_INPUT;
                                Uapi.uapiMessages.AppendText("Enabling Input\n", Color.Blue);
                                wrInfo = new CyGattWriteInfo(Globals.uapiHidData.inputs[Globals.uapiHidData.inputIdx].cccHdl, BitConverter.GetBytes((ushort)Constants.cccValues.ATT_CLIENT_CFG_NOTIFY));
                                err = Program.mf.gattClient.WriteCharacteristicWithoutResponse(wrInfo);
                                if (err.IsNotOK)
                                {
#if (debugTestHandlers)
                                    // The following is debug-only nonsense
                                    Utilities.ForceException(33);
#endif
                                }
                                return;
                            }
                        }
                    }
                    else
                    {
                        // TODO - handle any other possible characteristic reads...?
                        return;
                    }
                    Globals.hidState = Constants.hidStates.HID_STATE_INCOMPLETE;
                    Uapi.uapiMessages.AppendText("Invalid HID Report data!!\n", Color.HotPink);
                }
                else
                {
#if (debugTestHandlers)
                    // The following is debug-only nonsense
                    Utilities.ForceException(34);
#endif
                    string sm = string.Format("GATT read characteristic error: status = {0}",
                                              status.ToString());
                    Utilities.SetText(Program.mf.statusMessage, sm);
                }
            };

#if (newFunc)
            // Reports the value of characteristics with a specific UUID
            Program.mf.gattCb.CharacteristicReadByUUIDHandler = (result, status) =>
            {
            };

            // Reports the value read by ICyGattClient.ReadMultipleCharacteristic()
            Program.mf.gattCb.ReadMultipleCharacteristicsHandler = (result, status) =>
            {
            };
#endif
            // Reports the status of a characteristic value write
            Program.mf.gattCb.CharacteristicWriteHandler = (result, status) =>
            {
                string sm;

                if (status == CyStatus.BLE_STATUS_OK)
                {
                    if (Globals.hidState == Constants.hidStates.HID_STATE_ENABLING_INPUT)
                    {
                        if (result.Handle == Globals.uapiHidData.inputs[Globals.uapiHidData.inputIdx].cccHdl)
                        {
                            Globals.hidState = Constants.hidStates.HID_STATE_COMPLETE;
                            Uapi.uapiMessages.AppendText("UAPI Ready\n", Color.Blue);
                            return;
                        }
                    }
                    else if (Globals.hidState == Constants.hidStates.HID_STATE_COMPLETE)
                    {
                        if (result.Handle == Globals.uapiHidData.outputs[Globals.uapiHidData.outputIdx].dataHdl)
                        {
                            if (Uapi.txFragCnt != 0)
                            {
                                Uapi.SendMessage();
                            }
                        }
                        else
                        {
                            sm = string.Format("Unexpected write to handle {0}\n", result.Handle.ToString());
                            Uapi.uapiMessages.AppendText(sm, Color.Gold);
                        }
                        return;
                    }
                    Globals.hidState = Constants.hidStates.HID_STATE_INCOMPLETE;
                    Uapi.uapiMessages.AppendText("Invalid HID Report data!!\n", Color.HotPink);
                }
                else
                {
#if (debugTestHandlers)
                    // The following is debug-only nonsense
                    Utilities.ForceException(35);
#endif
                    sm = string.Format("GATT write characteristic error: status = {0}", status.ToString());
                    Utilities.SetText(Program.mf.statusMessage, sm);
                }
            };

#if (newFunc)
            // Reports the result of a reliable write
            Program.mf.gattCb.ReliableWriteCompletedHandler = (status) =>
            {
            };

            // Invoked when the peer device reports that the write buffer is full
            Program.mf.gattCb.WriteBufferFullHandler = (response) =>
            {
            };

            // Reports the descriptor value that was read
            Program.mf.gattCb.DescriptorReadHandler = (result, status) =>
            {
            };

            // Reports the status of a descriptor value write
            Program.mf.gattCb.DescriptorWriteHandler = (result, status) =>
            {
            };

            // Reports the status of ICyGattClient.GattStop()
            Program.mf.gattCb.GattStopHandler = (status) =>
            {
            };

            // Reports the discovered primary services
            Program.mf.discPrimCb.PrimaryServiceDiscoveredHandler = (services, status) =>
            {
            };

            // Reports the discovered included services
            Program.mf.findIncCb.FindIncludedServicesHandler = (result, status) =>
            {
            };

            // Reports the discovered characteristics
            Program.mf.discCharCb.CharacteristicsDiscoveredHandler = (result, status) =>
            {
            };

            // Reports the discovered descriptors
            Program.mf.discDescCb.DescriptorDiscoveredHandler = (result, status) =>
            {
            };
#endif
        }

        public static void SetupL2CapMgrHandlers()
        {
#if (newFunc)
            // Reports an L2CAP channel connection request received from a peer device
            Program.mf.l2capMgrCb.ChannelConnectionIndicationHandler = (response) =>
            {
            };

            // Reports an L2CAP channel disconnect initiated by the local device stack or by the peer device
            Program.mf.l2capMgrCb.ChannelDisconnectIndicationHandler = (info) =>
            {
            };

            // Reports the status of ICyL2CapMgr.DisconnectChannel()
            Program.mf.l2capMgrCb.DisconnectChannelHandler = (result, status) =>
            {
            };

            // Reports the established L2CAP channel
            Program.mf.l2capMgrCb.ChannelEstablishedHandler = (channel, status) =>
            {
            };

            // Reports the data received over an L2CAP channel
            Program.mf.l2capMgrCb.DataReceivedHandler = (info) =>
            {
            };

            // Reports that the receive credit of an L2CAP channel is low
            Program.mf.l2capMgrCb.ReceiveCreditLowIndicationHandler = (info) =>
            {
            };

            // Reports the transmit flow credit received from a peer device, for an L2CAP channel
            Program.mf.l2capMgrCb.TransmitCreditIndicationHandler = (info) =>
            {
            };

            // Reports the status of ICyL2CapMgr.SendData()
            Program.mf.l2capMgrCb.SendDataHandler = (channelID, status) =>
            {
            };

            // Reports the status of ICyL2CapMgr.SendCredits()
            Program.mf.l2capMgrCb.SendCreditsHandler = (channelID, status) =>
            {
            };
#endif
        }

        public static void StartScan()
        {
            CyApiErr err;
            string sm;

            // Setup the scan status changed handler
            Program.mf.scanCb.ScanStatusChangedHandler = (status) =>
            {
#if (debugTestHandlers)
                // The following is debug-only nonsense
                if (status != CyScanStatus.IN_PROGRESS)
                {
                    Utilities.ForceException(0);
                }
#endif
                sm = string.Format("Scan status: {0}", status);
                Utilities.SetText(Program.mf.statusMessage, sm);
                Utilities.SetState(Constants.hostStates.HOST_STATE_SCANNING);
            };

            // Using default settings
            err = Program.mf.bleMgr.StartScan(new CyBleScanSettings(), Program.mf.scanCb);
#if (debugTestHandlers)
            // The following is debug-only nonsense
            if (err.IsNotOK)
            {
                Utilities.ForceException(15);
            }
#endif
            Utilities.SetState(Constants.hostStates.HOST_STATE_SCAN_START);
        }

        public static void StopScan(bool wantToConnect)
        {
            CyApiErr err;
            string sm;

            // Setup the scan status changed handler
            Program.mf.scanCb.ScanStatusChangedHandler = (status) =>
            {
#if (debugTestHandlers)
                // The following is debug-only nonsense
                if (status != CyScanStatus.STOPPED)
                {
                    Utilities.ForceException(1);
                }
#endif
                sm = string.Format("Scan status: {0}", status);
                Utilities.SetText(Program.mf.statusMessage, sm);
                if (Globals.state == Constants.hostStates.HOST_STATE_SCAN_STOP_THEN_CONNECT)
                {
                    Program.mf.ConnectToDevice();
                }
                else
                {
                    Utilities.SetState(Constants.hostStates.HOST_STATE_READY);
                }
            };

            err = Program.mf.bleMgr.StopScan();
#if (debugTestHandlers)
            // The following is debug-only nonsense
            if (err.IsNotOK)
            {
                Utilities.ForceException(2);
            }
#endif
            if (wantToConnect)
            {
                Utilities.SetState(Constants.hostStates.HOST_STATE_SCAN_STOP_THEN_CONNECT);
            }
            else
            {
                Utilities.SetState(Constants.hostStates.HOST_STATE_SCAN_STOP);
            }
        }
    }
}
