﻿#undef debugTest
#define debugTest

using CySmart.Common.Base.Compatibility.Net20;
using CySmart.DongleCommunicator.API;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO.Ports;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;


namespace Host
{
    public partial class mainWindow : Form
    {
        private string[] ports;
        private string activePort;
        private int portIndex;

        private ushort minConnectionInterval;
        private ushort maxConnectionInterval;
        public ushort connectionInterval;
        public ushort slaveLatency;
        public ushort supervisionTimeout;
        private int connectionTimeout;

        public int connectionTimeLeft;

        public mainWindow()
        {
            InitializeComponent();
            communicator = null;
            bleMgrCb = null;
            activePort = null;
            portIndex = -1;
            ports = SerialPort.GetPortNames();
            minConnectionInterval = 8;
            maxConnectionInterval = 8;
            slaveLatency = 199;
            supervisionTimeout = 800;
            connectionTimeout = 15;
            dongleMgr = CySmartDongleMgr.GetInstance(); // Why can't we use 'new' here??
            Globals.hidState = Constants.hidStates.HID_STATE_INCOMPLETE;
            gattData = null;
        }

        private void mainWindow_Shown(object sender, EventArgs e)
        {
            Utilities.SetState(Constants.hostStates.HOST_STATE_IDLE);
            ConnectToDongle();
            this.Refresh();
        }

        private void mainWindow_FormClosing(object sender, FormClosingEventArgs e)
        {
            // TODO - add device disconnection here
            if (Globals.state == Constants.hostStates.HOST_STATE_SCANNING)
            {
                BleHandlers.StopScan(false);
            }
            if (portIndex != -1)
            {
                dongleMgr.CloseCommunicator(communicator);
            }
        }

        public void ConfigureButtons(bool dongleOn, string dongleText, bool scanOn, string scanText,
                                     bool deviceOn, string deviceText, bool gattOn, bool uapiOn)
        {
            Utilities.SetText(dongleButton, dongleText);
            Utilities.EnableButton(dongleButton, dongleOn);

            Utilities.SetText(scanButton, scanText);
            Utilities.EnableButton(scanButton, scanOn);

            Utilities.SetText(deviceButton, deviceText);
            Utilities.EnableButton(deviceButton, deviceOn);

            Utilities.EnableButton(gattButton, gattOn);

            Utilities.EnableButton(uapiButton, uapiOn);
        }

        private void dongleButton_Click(object sender, EventArgs e)
        {
            if (Globals.state == Constants.hostStates.HOST_STATE_IDLE)
            {
                ConnectToDongle();
            }
            else if (Globals.state == Constants.hostStates.HOST_STATE_READY)
            {
                CyApiErr err = dongleMgr.CloseCommunicator(communicator);
                if (err.IsOK)
                {
                    Utilities.SetText(statusMessage, "No dongle connected...");
                    Utilities.SetState(Constants.hostStates.HOST_STATE_IDLE);
                }
                else
                {
#if (debugTest)
                    // The following is debug-only nonsense
                    Utilities.ForceException(14);
#endif
                }
            }
        }

        private void scanButton_Click(object sender, EventArgs e)
        {
            if (Globals.state == Constants.hostStates.HOST_STATE_SCANNING)
            {
                BleHandlers.StopScan(false);
            }
            else if (Globals.state == Constants.hostStates.HOST_STATE_READY)
            {
                Utilities.ClearListView(scanResults);
                BleHandlers.StartScan();
            }
        }

        private void deviceButton_Click(object sender, EventArgs e)
        {
            if ((Globals.state == Constants.hostStates.HOST_STATE_READY) ||
                 (Globals.state == Constants.hostStates.HOST_STATE_SCANNING))
            {
                PrepareToConnectToDevice();
            }
            else if (Globals.state == Constants.hostStates.HOST_STATE_CONNECTING_DEVICE)
            {
                CancelDeviceConnection();
            }
            else if (Globals.state == Constants.hostStates.HOST_STATE_DEVICE_CONNECTED)
            {
                DisconnectDevice();
            }
        }

        private void gattButton_Click(object sender, EventArgs e)
        {
            if (Globals.state == Constants.hostStates.HOST_STATE_DEVICE_CONNECTED)
            {
                Gatt.gattDialogBox();
            }
        }

        private void uapiButton_Click(object sender, EventArgs e)
        {
            if (Globals.state == Constants.hostStates.HOST_STATE_DEVICE_CONNECTED)
            {
                Uapi.uapiDialogBox();
            }
        }

        private void scanResults_DoubleClick(object sender, EventArgs e)
        {
            if ((Globals.state == Constants.hostStates.HOST_STATE_READY) ||
                 (Globals.state == Constants.hostStates.HOST_STATE_SCANNING))
            {
                PrepareToConnectToDevice();
            }
        }

        private bool resizing = false;
        private void scanResults_SizeChanged(object sender, EventArgs e)
        {
            // Don't allow overlapping of SizeChanged calls
            if (!resizing)
            {
                // Set the resizing flag
                resizing = true;

                ListView listView = sender as ListView;
                if (listView != null)
                {
                    float totalColumnWidth = 0;

                    // Get the sum of all column tags
                    for (int i = 0; i < listView.Columns.Count; i++)
                    {
                        totalColumnWidth += Convert.ToInt32(listView.Columns[i].Tag);
                    }

                    // Calculate the percentage of space each column should 
                    // occupy in reference to the other columns and then set the 
                    // width of the column to that percentage of the visible space.
                    for (int i = 0; i < listView.Columns.Count; i++)
                    {
                        float colPercentage = (Convert.ToInt32(listView.Columns[i].Tag) / totalColumnWidth);
                        listView.Columns[i].Width = (int)(colPercentage * listView.ClientRectangle.Width + 0.5);
                    }
                }
            }

            // Clear the resizing flag
            resizing = false;
        }

        private void scanResults_SelectedIndexChanged(object sender, EventArgs e)
        {
            if ((Globals.state == Constants.hostStates.HOST_STATE_READY) ||
                 (Globals.state == Constants.hostStates.HOST_STATE_SCANNING))
            {
                Utilities.EnableButton(deviceButton, Utilities.IsConnectable(scanResults));
            }
            /*
            string sm;
            if (scanResults.SelectedItems.Count > 0)
            {
                sm = string.Format("Index = {0}", scanResults.SelectedIndices[0]);
            }
            else
            {
                sm = string.Format("Index = <none>");

            }
            SetText(statusMessage, sm);
            */
        }

        private void connectionTimer_Tick(object sender, EventArgs e)
        {
            if (connectionTimeLeft > 0)
            {
                --connectionTimeLeft;
            }
            else
            {
                CancelDeviceConnection();
            }
        }

        private int DongleDialogBox()
        {
            Form prompt = new Form();
            prompt.Width = 500;
            prompt.Height = 200;
            prompt.Text = "Select BLE Dongle COM Port";
            prompt.Icon = this.Icon;

            Label textLabel = new Label()
            {
                Left = 50,
                Top = 20,
                Text = "Active COM Ports:"
            };
            textLabel.Height = 20;

            ComboBox activePorts = new ComboBox()
            {
                Left = 50,
                Top = 40,
                Width = 400
            };
            foreach (string port in ports)
            {
                activePorts.Items.Add(port);
            }
            if (portIndex == -1)
            {
                activePorts.SelectedIndex = 0;
            }
            else
            {
                activePorts.SelectedIndex = portIndex;
            }

            Button ok = new Button()
            {
                Text = "Ok",
                Left = 50,
                Width = 100,
                Top = 70
            };
            ok.Click += (sender, e) =>
            {
                prompt.Close();
                prompt.Dispose();
            };

            Button cancel = new Button()
            {
                Text = "Cancel",
                Left = 350,
                Width = 100,
                Top = 70
            };
            cancel.Click += (sender, e) =>
            {
                activePorts.SelectedIndex = -1;
                prompt.Close();
                prompt.Dispose();
            };

            prompt.Controls.Add(ok);
            prompt.Controls.Add(cancel);
            prompt.Controls.Add(textLabel);
            prompt.Controls.Add(activePorts);

            prompt.ShowDialog();

            return activePorts.SelectedIndex;
        }

        private CyBleConnectionSettings ConnectionDialogBox()
        {
            CyBleConnectionSettings settings = null;

            Form prompt = new Form();
            prompt.Width = 500;
            prompt.Height = 400;
            prompt.Text = "Enter BLE connection parameters";
            prompt.Icon = this.Icon;

            Label minIntervalLabel = new Label()
            {
                Left = 50,
                Top = 20,
                Text = "Minimum Connection Interval:",
                Height = 20,
                Width = 150
            };
            NumericUpDown minInterval = new NumericUpDown()
            {
                Left = 50,
                Top = 40,
                Width = 150,
                Minimum = 6,
                Maximum = maxConnectionInterval,
                Value = minConnectionInterval
            };

            Label maxIntervalLabel = new Label()
            {
                Left = 275,
                Top = 20,
                Text = "Maximum Connection Interval:",
                Height = 20,
                Width = 150
            };
            NumericUpDown maxInterval = new NumericUpDown()
            {
                Left = 275,
                Top = 40,
                Width = 150,
                Minimum = minConnectionInterval,
                Maximum = 3200,
                Value = maxConnectionInterval
            };

            minInterval.ValueChanged += (sender, e) =>
            {
                maxInterval.Minimum = minInterval.Value;
            };
            maxInterval.ValueChanged += (sender, e) =>
            {
                minInterval.Maximum = maxInterval.Value;
            };

            Label latencyLabel = new Label()
            {
                Left = 50,
                Top = 80,
                Text = "Slave Latency:",
                Height = 20,
                Width = 150
            };
            NumericUpDown latency = new NumericUpDown()
            {
                Left = 50,
                Top = 100,
                Width = 150,
                Minimum = 0,
                Maximum = 3200,
                Value = slaveLatency
            };

            Label superTimeoutLabel = new Label()
            {
                Left = 275,
                Top = 80,
                Text = "Supervision Timeout:",
                Height = 20,
                Width = 150
            };
            NumericUpDown superTimeout = new NumericUpDown()
            {
                Left = 275,
                Top = 100,
                Width = 150,
                Minimum = 10,
                Maximum = 3200,
                Value = supervisionTimeout
            };

            Label connTimeoutLabel = new Label()
            {
                Left = 50,
                Top = 140,
                Text = "Connection Timeout:",
                Height = 20,
                Width = 150
            };
            NumericUpDown connTimeout = new NumericUpDown()
            {
                Left = 50,
                Top = 160,
                Width = 150,
                Minimum = 3,
                Maximum = 30,
                Value = connectionTimeout
            };

            Button ok = new Button()
            {
                Text = "Ok",
                Left = 50,
                Width = 100,
                Top = 300
            };
            ok.Click += (sender, e) =>
            {
                minConnectionInterval = (ushort)minInterval.Value;
                maxConnectionInterval = (ushort)maxInterval.Value;
                slaveLatency = (ushort)latency.Value;
                supervisionTimeout = (ushort)superTimeout.Value;
                connectionTimeout = (ushort)connTimeout.Value;
                settings = new CyBleConnectionSettings(minConnectionInterval,
                                                       maxConnectionInterval,
                                                       CyBleConnectionSettings.DEFAULT_MINIMUM_CE_LENGTH,
                                                       CyBleConnectionSettings.DEFAULT_MAXIMUM_CE_LENGTH,
                                                       CyBleConnectionSettings.DEFAULT_INITIATOR_FILTER_POLICY,
                                                       CyBleConnectionSettings.DEFAULT_INITIATOR_ADDRESS_TYPE,
                                                       slaveLatency,
                                                       supervisionTimeout,
                                                       CyBleConnectionSettings.DEFAULT_SCAN_INTERVAL,
                                                       CyBleConnectionSettings.DEFAULT_SCAN_WINDOW);
                prompt.Close();
                prompt.Dispose();
            };

            Button cancel = new Button()
            {
                Text = "Cancel",
                Left = 325,
                Width = 100,
                Top = 300
            };
            cancel.Click += (sender, e) =>
            {
                settings = null;
                prompt.Close();
                prompt.Dispose();
            };

            prompt.Controls.Add(ok);
            prompt.Controls.Add(cancel);

            prompt.Controls.Add(minIntervalLabel);
            prompt.Controls.Add(maxIntervalLabel);
            prompt.Controls.Add(latencyLabel);
            prompt.Controls.Add(superTimeoutLabel);
            prompt.Controls.Add(connTimeoutLabel);

            prompt.Controls.Add(minInterval);
            prompt.Controls.Add(maxInterval);
            prompt.Controls.Add(latency);
            prompt.Controls.Add(superTimeout);
            prompt.Controls.Add(connTimeout);

            prompt.ShowDialog();

            return settings;
        }

        // Connect to CySmart BLE dongle
        private void ConnectToDongle()
        {
            CyApiErr err;
            string sm;

            Utilities.SetState(Constants.hostStates.HOST_STATE_CONNECTING_DONGLE);

            // Prompt the user to select a COM port, and then either attempt to connect to the dongle on that port
            //  or cancel the connection attempt
            do
            {
                portIndex = DongleDialogBox();
                if (portIndex == -1)
                {
                    // Here, the user clicked on the 'Close' button
                    break;
                }
                activePort = ports[portIndex];
                sm = string.Format("Connecting to {0}", activePort);
                Utilities.SetText(statusMessage, sm);
                err = dongleMgr.TryGetCySmartDongleCommunicator(new CyDongleInfo(activePort, CyDongleInfo.CySmartDongleType.CY5677),
                                                                out communicator);
                if (err.IsNotOK)
                {
                    sm = string.Format("{0} Connection Failed!", activePort);
                    Utilities.SetText(statusMessage, sm);
                }
            } while (err.IsNotOK);

            if (portIndex == -1)
            {
                Utilities.SetText(statusMessage, "No dongle connected...");
                Utilities.SetState(Constants.hostStates.HOST_STATE_IDLE);
            }
            else
            {
                sm = string.Format("Connected to {0} dongle", activePort);
                Utilities.SetText(statusMessage, sm);

                bleMgrCb = new bleMgrCallback();
                err = bleMgr.RegisterBleMgrCallback(bleMgrCb);

                // scanCb is passed to CySmart when scanning is started, so no separate registration req'd
                scanCb = new scanCallback();

                devListCb = new deviceListCallback();
                err = devList.RegisterDeviceListCallback(devListCb);

                secMgrCb = new securityMgrCallback();
                err = secMgr.RegisterSecurityMgrCallback(secMgrCb);

                devAddrMgrCb = new deviceAddressMgrCallback();
                err = devAddrMgr.RegisterDeviceAddressMgrCallback(devAddrMgrCb);

                // These next callbacks won't be registered until after a remote device has connected to us,
                //  but we instantiate them now so that we can set up their handlers in advance
                devCb = new deviceCallback();
                gattCb = new gattClientCallback();
                discPrimCb = new discoverPrimaryServiceCallback();
                findIncCb = new findIncludedServicesCallback();
                discCharCb = new discoverCharacteristicsCallback();
                discDescCb = new discoverDescriptorsCallback();
                l2capMgrCb = new l2capMgrCallback();

                BleHandlers.SetupBleMgrHandlers();              // For bleMgrCb
                BleHandlers.SetupScanHandlers();                // For scanCb
                BleHandlers.SetupDeviceListHandlers();          // For devListCb
                BleHandlers.SetupSecurityMgrHandlers();         // For secMgrCb
                BleHandlers.SetupDeviceAddressMgrHandlers();    // For devAddrMgrCb
                BleHandlers.SetupDeviceHandlers();              // For devCb
                BleHandlers.SetupGattHandlers();                // For gattCb, discPrimCb, findIncCb, discCharCb, and discDescCb
                BleHandlers.SetupL2CapMgrHandlers();            // For l2capMgrCb

                Utilities.SetState(Constants.hostStates.HOST_STATE_READY);
            }
        }

        private void PrepareToConnectToDevice()
        {
            if (Utilities.IsConnectable(scanResults))
            {
                // Save the currently highlighted device, in case the highlight changes after we've already
                //  started the connection process
                device = highlightedDevice;
                if (Globals.state == Constants.hostStates.HOST_STATE_SCANNING)
                {
                    BleHandlers.StopScan(true);
                }
                else    // Here, Globals.state is HOST_STATE_READY
                {
                    ConnectToDevice();
                }
            }
        }

        public void ConnectToDevice()
        {
            CyApiErr err;

            Utilities.SetState(Constants.hostStates.HOST_STATE_CONNECTION_PENDING);

            CyBleConnectionSettings connectionSettings = ConnectionDialogBox();

            if (connectionSettings == null)
            {
                Utilities.SetState(Constants.hostStates.HOST_STATE_READY);
            }
            else
            {
                err = bleMgr.Connect(new CyConnectInfo(device.PeerDeviceAddress, connectionSettings));
                if (err.IsNotOK)
                {
#if (debugTest)
                    Utilities.ForceException(5);
#endif
                }
                string sm = string.Format("Connecting to {0:X12}", device.PeerDeviceAddress.Address);
                Utilities.SetText(statusMessage, sm);
                connectionTimeLeft = connectionTimeout;
                connectionTimer.Start();
                Utilities.SetState(Constants.hostStates.HOST_STATE_CONNECTING_DEVICE);
            }
        }

        private void CancelDeviceConnection()
        {
            connectionTimer.Stop();
            CyApiErr err = bleMgr.CancelConnection(device.PeerDeviceAddress);
            if (err.IsNotOK)
            {
#if (debugTest)
                Utilities.ForceException(5);
#endif
            }
            Utilities.SetState(Constants.hostStates.HOST_STATE_CANCELING_CONNECTION);
        }

        private void DisconnectDevice()
        {
            CyApiErr err = bleMgr.Disconnect(connectedDevice);
            if (err.IsNotOK)
            {
#if (debugTest)
                Utilities.ForceException(6);
#endif
            }
            Utilities.SetState(Constants.hostStates.HOST_STATE_DISCONNECTING);
        }
    }
}
