﻿#undef debugTestConstants
#define debugTestConstants

using CySmart.Common.Base.Compatibility.Net20;
using CySmart.DongleCommunicator.API;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO.Ports;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;


namespace Host
{
    public sealed class Constants
    {
        static Constants()
        {
            // Could this explicit initialization have been done where 'profiles' was declared below?
            profiles = new Dictionary<byte, profileFlags>()
            {
                { 0x01, profileFlags.ZRC11 },
                { 0x02, profileFlags.ZID },
                { 0x03, profileFlags.ZRC20 },
                { 0xC0, profileFlags.MSO },
                { 0xC1, profileFlags.MSO_VOICE },
                { 0xC2, profileFlags.MSO_OTA },
                { 0xE0, profileFlags.ESP }
            };
        }

        public enum hostStates
        {
            HOST_STATE_IDLE,
            HOST_STATE_CONNECTING_DONGLE,
            HOST_STATE_READY,
            HOST_STATE_SCAN_START,
            HOST_STATE_SCANNING,
            HOST_STATE_SCAN_STOP,
            HOST_STATE_SCAN_STOP_THEN_CONNECT,
            HOST_STATE_CONNECTION_PENDING,
            HOST_STATE_CONNECTING_DEVICE,
            HOST_STATE_CANCELING_CONNECTION,
            HOST_STATE_PAIRING,
            HOST_STATE_DISCOVERING_GATT,
            HOST_STATE_DEVICE_CONNECTED,
            HOST_STATE_GATT_SHOWN,
            HOST_STATE_UAPI_SHOWN,
            HOST_STATE_DISCONNECTING
        };

        public enum hidStates
        {
            HID_STATE_INCOMPLETE,
            HID_STATE_CONFIRMING_INPUT,
            HID_STATE_CONFIRMING_OUTPUT,
            HID_STATE_ENABLING_INPUT,
            HID_STATE_COMPLETE
        };

        public enum configParameterAccess : byte
        {
            NONE = 0x00,
            READ = 0x01,
            WRITE = 0x10
        }

        public enum profileFlags : byte
        {
            ZRC11       = 0x01,
            ZID         = 0x02,
            ZRC20       = 0x04,
            MSO         = 0x08,
            MSO_VOICE   = 0x10,
            MSO_OTA     = 0x20,
            ESP         = 0x40,
            ZRC11_ZRC20         = ZRC11 | ZRC20,
            ZRC20_MSO           = ZRC20 | MSO,
            MSO_MSO_VOICE       = MSO | MSO_VOICE,
            MSO_MSO_OTA         = MSO | MSO_OTA,
            MSO_MSO_VOICE_OTA   = MSO | MSO_VOICE | MSO_OTA,
            COMMON              = ZRC11 | ZID | ZRC20 | MSO | ESP
        }

        public static Dictionary<byte, profileFlags> profiles;

        public enum uapiOptions : byte
        {
            NONE = 0x00,
            RIR = 0x04,
            MORE_DATA = 0x08,
            START_STOP_FRAG = 0x10,
            FRAG_CONTINUE = 0x20,
            ACK_RESPONSE = 0x40,
            ACK_REQUEST = 0x80
        }

        public enum cccValues : ushort
        {
            ATT_CLIENT_CFG_NOTIFY = 0x0001,
            ATT_CLIENT_CFG_INDICATE = 0x0002
        }

        public enum hidReportTypes : byte
        {
            HID_REPORT_TYPE_INPUT = 0x01,
            HID_REPORT_TYPE_OUTPUT = 0x02,
            HID_REPORT_TYPE_FEATURE = 0x03
        }

        public enum hidReportIDs : byte
        {
            HID_REPORT_ID_UAPI_IN = 0x07,
            HID_REPORT_ID_UAPI_OUT = 0x08,
            HID_REPORT_ID_VOICE_IN = 0x1F
        }

        public enum uuids : ushort
        {
            // Service UUIDs
            ATT_UUID_GAP_SERVICE = 0x1800,   // Generic Access Profile Service
            ATT_UUID_GATT_SERVICE = 0x1801,   // Generic Attribute Profile Service
            ATT_UUID_IMMEDIATE_ALERT_SERVICE = 0x1802,   // Immediate Alert Service
            ATT_UUID_LINK_LOSS_SERVICE = 0x1803,   // Link Loss Service
            ATT_UUID_TX_POWER_SERVICE = 0x1804,   // Tx Power Service
            ATT_UUID_CURRENT_TIME_SERVICE = 0x1805,   // Current Time Service
            ATT_UUID_REF_TIME_UPDATE_SERVICE = 0x1806,   // Reference Time Update Service
            ATT_UUID_DST_CHANGE_SERVICE = 0x1807,   // Next DST Change Service
            ATT_UUID_GLUCOSE_SERVICE = 0x1808,   // Glucose Service
            ATT_UUID_HEALTH_THERM_SERVICE = 0x1809,   // Health Thermometer Service
            ATT_UUID_DEVICE_INFO_SERVICE = 0x180A,   // Device Information Service
            ATT_UUID_NETWORK_AVAIL_SERVICE = 0x180B,   // Network Availability Service
            ATT_UUID_WATCHDOG_SERVICE = 0x180C,   // Watchdog Service
            ATT_UUID_HEART_RATE_SERVICE = 0x180D,   // Heart Rate Service
            ATT_UUID_PHONE_ALERT_SERVICE = 0x180E,   // Phone Alert Status Service
            ATT_UUID_BATTERY_SERVICE = 0x180F,   // Battery Service
            ATT_UUID_BLOOD_PRESSURE_SERVICE = 0x1810,   // Blood Pressure Service
            ATT_UUID_ALERT_NOTIF_SERVICE = 0x1811,   // Alert Notification Service
            ATT_UUID_HID_SERVICE = 0x1812,   // Human Interface Device Service
            ATT_UUID_SCAN_PARAM_SERVICE = 0x1813,   // Scan Parameter Service
            ATT_UUID_RUNNING_SPEED_SERVICE = 0x1814,   // Running Speed Service
            ATT_UUID_CYCLING_SPEED_SERVICE = 0x1816,   // Cycling Speed Service
            ATT_UUID_CYCLING_POWER_SERVICE = 0x1818,   // Cycling Power Service
            ATT_UUID_USER_DATA_SERVICE = 0x181C,   // User Data Service
            ATT_UUID_IP_SUPPORT_SERVICE = 0x1820,   // IP Support Service
            ATT_UUID_PULSE_OXIMITER_SERVICE = 0x1822,   // Pulse Oximeter Service

            // GATT UUIDs
            ATT_UUID_PRIMARY_SERVICE = 0x2800,   // Primary Service
            ATT_UUID_SECONDARY_SERVICE = 0x2801,   // Secondary Service
            ATT_UUID_INCLUDE = 0x2802,   // Include
            ATT_UUID_CHARACTERISTIC = 0x2803,   // Characteristic

            // Descriptor UUIDs
            ATT_UUID_CHARACTERISTIC_EXT = 0x2900,   // Characteristic Extended Properties
            ATT_UUID_CHAR_USER_DESC = 0x2901,   // Characteristic User Description
            ATT_UUID_CLIENT_CHAR_CONFIG = 0x2902,   // Client Characteristic Configuration
            ATT_UUID_SERVER_CHAR_CONFIG = 0x2903,   // Server Characteristic Configuration
            ATT_UUID_CHAR_PRES_FORMAT = 0x2904,   // Characteristic Presentation Format
            ATT_UUID_AGGREGATE_FORMAT = 0x2905,   // Characteristic Aggregate Format
            ATT_UUID_VALID_RANGE = 0x2906,   // Valid Range
            ATT_UUID_HID_EXT_REPORT_MAPPING = 0x2907,   // HID External Report ID Mapping
            ATT_UUID_HID_REPORT_ID_MAPPING = 0x2908,   // HID Report ID Mapping

            // Characteristic UUIDs
            ATT_UUID_DEVICE_NAME = 0x2A00,   // Device Name
            ATT_UUID_APPEARANCE = 0x2A01,   // Appearance
            ATT_UUID_PERIPH_PRIVACY_FLAG = 0x2A02,   // Peripheral Privacy Flag
            ATT_UUID_RECONN_ADDR = 0x2A03,   // Reconnection Address
            ATT_UUID_PREF_CONN_PARAM = 0x2A04,   // Peripheral Preferred Connection Parameters
            ATT_UUID_SERVICE_CHANGED = 0x2A05,   // Service Changed
            ATT_UUID_ALERT_LEVEL = 0x2A06,   // Alert Level
            ATT_UUID_TX_POWER_LEVEL = 0x2A07,   // Tx Power Level
            ATT_UUID_DATE_TIME = 0x2A08,   // Date Time
            ATT_UUID_DAY_OF_WEEK = 0x2A09,   // Day of Week
            ATT_UUID_DAY_DATE_TIME = 0x2A0A,   // Day Date Time
            ATT_UUID_EXACT_TIME_100 = 0x2A0B,   // Exact Time 100
            ATT_UUID_EXACT_TIME_256 = 0x2A0C,   // Exact Time 256
            ATT_UUID_DST_OFFSET = 0x2A0D,   // DST Offset
            ATT_UUID_TIME_ZONE = 0x2A0E,   // Time Zone
            ATT_UUID_LOCAL_TIME_INFO = 0x2A0F,   // Local Time Information
            ATT_UUID_SECONDARY_TIME_ZONE = 0x2A10,   // Secondary Time Zone
            ATT_UUID_TIME_WITH_DST = 0x2A11,   // Time with DST
            ATT_UUID_TIME_ACCURACY = 0x2A12,   // Time Accuracy
            ATT_UUID_TIME_SOURCE = 0x2A13,   // Time Source
            ATT_UUID_REFERENCE_TIME_INFO = 0x2A14,   // Reference Time Information
            ATT_UUID_TIME_BROADCAST = 0x2A15,   // Time Broadcast
            ATT_UUID_TIME_UPDATE_CP = 0x2A16,   // Time Update Control Point
            ATT_UUID_TIME_UPDATE_STATE = 0x2A17,   // Time Update State
            ATT_UUID_GLUCOSE_MEAS = 0x2A18,   // Glucose Measurement
            ATT_UUID_BATTERY_LEVEL = 0x2A19,   // Battery Level
            ATT_UUID_BATTERY_POWER_STATE = 0x2A1A,   // Battery Power State
            ATT_UUID_BATTERY_LEVEL_STATE = 0x2A1B,   // Battery Level State
            ATT_UUID_TEMP_MEAS = 0x2A1C,   // Temperature Measurement
            ATT_UUID_TEMP_TYPE = 0x2A1D,   // Temperature Type
            ATT_UUID_INTERMEDIATE_TEMP = 0x2A1E,   // Intermediate Temperature
            ATT_UUID_TEMP_C = 0x2A1F,   // Temperature Celsius
            ATT_UUID_TEMP_F = 0x2A20,   // Temperature Fahrenheit
            ATT_UUID_MEAS_INTERVAL = 0x2A21,   // Measurement Interval
            ATT_UUID_HID_BOOT_KEYBOARD_IN = 0x2A22,   // HID Boot Keyboard In
            ATT_UUID_SYSTEM_ID = 0x2A23,   // System ID
            ATT_UUID_MODEL_NUMBER = 0x2A24,   // Model Number String
            ATT_UUID_SERIAL_NUMBER = 0x2A25,   // Serial Number String
            ATT_UUID_FIRMWARE_REV = 0x2A26,   // Firmware Revision String
            ATT_UUID_HARDWARE_REV = 0x2A27,   // Hardware Revision String
            ATT_UUID_SOFTWARE_REV = 0x2A28,   // Software Revision String
            ATT_UUID_MANUFACTURER_NAME = 0x2A29,   // Manufacturer Name String
            ATT_UUID_11073_CERT_DATA = 0x2A2A,   // IEEE 11073-20601 Regulatory Certification Data List
            ATT_UUID_CURRENT_TIME = 0x2A2B,   // Current Time
            ATT_UUID_ELEVATION = 0x2A2C,   // Elevation
            ATT_UUID_LATITUDE = 0x2A2D,   // Latitude
            ATT_UUID_LONGITUDE = 0x2A2E,   // Longitude
            ATT_UUID_POSITION_2D = 0x2A2F,   // Position 2D
            ATT_UUID_POSITION_3D = 0x2A30,   // Position 3D
            ATT_UUID_VENDOR_ID = 0x2A31,   // Vendor ID
            ATT_UUID_HID_BOOT_KEYBOARD_OUT = 0x2A32,   // HID Boot Keyboard Out
            ATT_UUID_HID_BOOT_MOUSE_IN = 0x2A33,   // HID Boot Mouse In
            ATT_UUID_GLUCOSE_MEAS_CONTEXT = 0x2A34,   // Glucose Measurement Context
            ATT_UUID_BP_MEAS = 0x2A35,   // Blood Pressure Measurement
            ATT_UUID_INTERMEDIATE_BP = 0x2A36,   // Intermediate Cuff Pressure
            ATT_UUID_HR_MEAS = 0x2A37,   // Heart Rate Measurement
            ATT_UUID_HR_SENSOR_LOC = 0x2A38,   // Body Sensor Location
            ATT_UUID_HR_CP = 0x2A39,   // Heart Rate Control Point
            ATT_UUID_REMOVABLE = 0x2A3A,   // Removable
            ATT_UUID_SERVICE_REQ = 0x2A3B,   // Service Required
            ATT_UUID_SCI_TEMP_C = 0x2A3C,   // Scientific Temperature in Celsius
            ATT_UUID_STRING = 0x2A3D,   // String
            ATT_UUID_NETWORK_AVAIL = 0x2A3E,   // Network Availability
            ATT_UUID_ALERT_STATUS = 0x2A3F,   // Alert Status
            ATT_UUID_RINGER_CP = 0x2A40,   // Ringer Control Point
            ATT_UUID_RINGER_SETTING = 0x2A41,   // Ringer Setting
            ATT_UUID_ALERT_CAT_ID_MASK = 0x2A42,   // Alert Category ID Bit Mask
            ATT_UUID_ALERT_CAT_ID = 0x2A43,   // Alert Category ID
            ATT_UUID_ALERT_NOTIF_CP = 0x2A44,   // Alert Notification Control Point
            ATT_UUID_UNREAD_ALERT_STATUS = 0x2A45,   // Unread Alert Status
            ATT_UUID_NEW_ALERT = 0x2A46,   // New Alert
            ATT_UUID_SUP_NEW_ALERT_CAT = 0x2A47,   // Supported New Alert Category
            ATT_UUID_SUP_UNREAD_ALERT_CAT = 0x2A48,   // Supported Unread Alert Category
            ATT_UUID_BP_FEATURE = 0x2A49,   // Blood Pressure Feature
            ATT_UUID_HID_INFORMATION = 0x2A4A,   // HID Information
            ATT_UUID_HID_REPORT_MAP = 0x2A4B,   // HID Report Map
            ATT_UUID_HID_CONTROL_POINT = 0x2A4C,   // HID Control Point
            ATT_UUID_HID_REPORT = 0x2A4D,   // HID Report
            ATT_UUID_HID_PROTOCOL_MODE = 0x2A4E,   // HID Protocol Mode
            ATT_UUID_SCAN_INT_WIND = 0x2A4F,   // Scan Interval Window
            ATT_UUID_PNP_ID = 0x2A50,   // PnP ID
            ATT_UUID_GLUCOSE_FEATURE = 0x2A51,   // Glucose Feature
            ATT_UUID_RACP = 0x2A52,   // Record Access Control Point
            ATT_UUID_RUNNING_SPEED_MEASUREMENT = 0x2A53,   // Running Speed Measurement
            ATT_UUID_RUNNING_SPEED_FEATURE = 0x2A54,   // Running Speed Feature
            ATT_UUID_CYCLING_SPEED_MEASUREMENT = 0x2A5B,   // Cycling Speed Measurement
            ATT_UUID_CYCLING_SPEED_FEATURE = 0x2A5C,   // Cycling Speed Feature
            ATT_UUID_SENSOR_LOCATION = 0x2A5D,   // Sensor Location
            ATT_UUID_PULSE_OX_SPOT_CHECK = 0x2A5E,   // Pulse Oximeter Spot Check
            ATT_UUID_PULSE_OX_CONTINUOUS = 0x2A5F,   // Pulse Oximeter Continuous
            ATT_UUID_PULSE_OX_FEATURES = 0x2A60,   // Pulse Oximeter Features
            ATT_UUID_CYCLING_POWER_MEASUREMENT = 0x2A63,   // Cycling Power Measurement
            ATT_UUID_CYCLING_POWER_FEATURE = 0x2A65,   // Cycling Power Feature
            ATT_UUID_DB_CHANGE_INCREMENT = 0x2A99,   // Database Change Increment
            ATT_UUID_USER_INDEX = 0x2A9A,   // User Index
            ATT_UUID_USER_CONTROL_POINT = 0x2A9F,   // User Control Point
            ATT_UUID_CAR = 0x2AA6,   // Central Address Resolution
            ATT_UUID_RPAO = 0x2AC9,   // Resolvable Private Address Only

            // Miscellaneous
            ATT_UUID_GENERIC_CTRL_SERVICE = 0xF011,   // Generic Control Service
            ATT_UUID_COMMAND_ENUM = 0xE010,   // Command Enumeration
            ATT_UUID_GENERIC_COMMAND_CP = 0xE011,   // Generic Command Control Point
            ATT_UUID_WEIGHT_SCALE_SERVICE = 0x181D,   // Weight Scale Service
            ATT_UUID_WEIGHT_MEAS = 0x2A9D,   // Weight Measurement
            ATT_UUID_WEIGHT_SCALE_FEATURE = 0x2A9E,   // Weight Scale Feature

            // Unit UUIDs
            ATT_UUID_UNITLESS = 0x2700,   // unitless
            ATT_UUID_LENGTH_M = 0x2701,   // length metre
            ATT_UUID_MASS_KG = 0x2702,   // mass kilogram
            ATT_UUID_TIME_SEC = 0x2703,   // time second
            ATT_UUID_ELECTRIC_CURRENT_AMP = 0x2704,   // electric current ampere
            ATT_UUID_THERMO_TEMP_K = 0x2705,   // thermodynamic temperature kelvin
            ATT_UUID_AMOUNT_OF_SUBSTANCE_MOLE = 0x2706,   // amount of substance mole
            ATT_UUID_LUMINOUS_INTENSITY_CAND = 0x2707,   // luminous intensity candela
            ATT_UUID_AREA_SQ_M = 0x2710,   // area square metres
            ATT_UUID_VOLUME_CU_M = 0x2711,   // volume cubic metres
            ATT_UUID_VELOCITY_MPS = 0x2712,   // velocity metres per second
            ATT_UUID_ACCELERATION_MPS_SQ = 0x2713,   // acceleration metres per second squared
            ATT_UUID_WAVENUMBER_RECIPROCAL_M = 0x2714,   // wavenumber reciprocal metre
            ATT_UUID_DENSITY_KG_PER_CU_M = 0x2715,   // density kilogram per cubic metre
            ATT_UUID_SURFACE_DENS_KG_PER_SQ_M = 0x2716,   // surface density kilogram per square metre
            ATT_UUID_SPECIFIC_VOL_CU_M_PER_KG = 0x2717,   // specific volume cubic metre per kilogram
            ATT_UUID_CURRENT_DENS_AMP_PER_SQ_M = 0x2718,   // current density ampere per square metre
            ATT_UUID_MAG_FIELD_STR_AMP_PER_M = 0x2719,   // magnetic field strength ampere per metre
            ATT_UUID_AMOUNT_CONC_MOLE_PER_CU_M = 0x271A,   // amount concentration mole per cubic metre
            ATT_UUID_MASS_CONC_KG_PER_CU_M = 0x271B,   // mass concentration kilogram per cubic metre
            ATT_UUID_LUM_CAND_PER_SQ_M = 0x271C,   // luminance candela per square metre
            ATT_UUID_REFRACTIVE_INDEX = 0x271D,   // refractive index
            ATT_UUID_RELATIVE_PERMEABILITY = 0x271E,   // relative permeability
            ATT_UUID_PLANE_ANGLE_R = 0x2720,   // plane angle radian
            ATT_UUID_SOLID_ANGLE_STER = 0x2721,   // solid angle steradian
            ATT_UUID_FREQUENCY_HERTZ = 0x2722,   // frequency hertz
            ATT_UUID_FORCE_NEWT = 0x2723,   // force newton
            ATT_UUID_PRESSURE_PASCAL = 0x2724,   // pressure pascal
            ATT_UUID_ENERGY_J = 0x2725,   // energy joule
            ATT_UUID_POWER_W = 0x2726,   // power watt
            ATT_UUID_ELECTRIC_CHG_C = 0x2727,   // electric charge coulomb
            ATT_UUID_ELECTRIC_POTENTIAL_VOLT = 0x2728,   // electric potential difference volt
            ATT_UUID_CAPACITANCE_F = 0x2729,   // capacitance farad
            ATT_UUID_ELECTRIC_RESISTANCE_OHM = 0x272A,   // electric resistance ohm
            ATT_UUID_ELECTRIC_COND_SIEMENS = 0x272B,   // electric conductance siemens
            ATT_UUID_MAGNETIC_FLEX_WEBER = 0x272C,   // magnetic flex weber
            ATT_UUID_MAGNETIC_FLEX_DENS_TESLA = 0x272D,   // magnetic flex density tesla
            ATT_UUID_INDUCTANCE_H = 0x272E,   // inductance henry
            ATT_UUID_C_TEMP_DEG_C = 0x272F,   // Celsius temperature degree Celsius
            ATT_UUID_LUMINOUS_FLUX_LUMEN = 0x2730,   // luminous flux lumen
            ATT_UUID_ILLUMINANCE_LUX = 0x2731,   // illuminance lux
            ATT_UUID_RADIONUCLIDE_BECQUEREL = 0x2732,   // activity referred to a radionuclide becquerel
            ATT_UUID_ABSORBED_DOSE_GRAY = 0x2733,   // absorbed dose gray
            ATT_UUID_DOSE_EQUIVALENT_SIEVERT = 0x2734,   // dose equivalent sievert
            ATT_UUID_CATALYTIC_ACTIVITY_KATAL = 0x2735,   // catalytic activity katal
            ATT_UUID_DYNAMIC_VISC_PASCAL_SEC = 0x2740,   // dynamic viscosity pascal second
            ATT_UUID_MOMENT_OF_FORCE_NEWT_M = 0x2741,   // moment of force newton metre
            ATT_UUID_SURFACE_TENSION_NEWT_PER_M = 0x2742,   // surface tension newton per metre
            ATT_UUID_ANG_VELOCITY_R_PER_SEC = 0x2743,   // angular velocity radian per second
            ATT_UUID_ANG_ACCEL_R_PER_SEC_SQD = 0x2744,   // angular acceleration radian per second squared
            ATT_UUID_HEAT_FLUX_DEN_W_PER_SQ_M = 0x2745,   // heat flux density watt per square metre
            ATT_UUID_HEAT_CAP_J_PER_K = 0x2746,   // heat capacity joule per kelvin
            ATT_UUID_SPEC_HEAT_CAP_J_PER_KG_K = 0x2747,   // specific heat capacity joule per kilogram kelvin
            ATT_UUID_SPEC_ENERGY_J_PER_KG = 0x2748,   // specific energy joule per kilogram
            ATT_UUID_THERMAL_COND_W_PER_M_K = 0x2749,   // thermal conductivity watt per metre kelvin
            ATT_UUID_ENERGY_DENSITY_J_PER_CU_M = 0x274A,   // energy density joule per cubic metre
            ATT_UUID_ELEC_FIELD_STR_VOLT_PER_M = 0x274B,   // electric field strength volt per metre
            ATT_UUID_ELEC_CHG_DENS_C_PER_CU_M = 0x274C,   // electric charge density coulomb per cubic metre
            ATT_UUID_SURF_CHG_DENS_C_PER_SQ_M = 0x274D,   // surface charge density coulomb per square metre
            ATT_UUID_ELEC_FLUX_DENS_C_PER_SQ_M = 0x274E,   // electric flux density coulomb per square metre
            ATT_UUID_PERMITTIVITY_F_PER_M = 0x274F,   // permittivity farad per metre
            ATT_UUID_PERMEABILITY_H_PER_M = 0x2750,   // permeability henry per metre
            ATT_UUID_MOLAR_ENERGY_J_PER_MOLE = 0x2751,   // molar energy joule per mole
            ATT_UUID_MOLAR_ENTROPY_J_PER_MOLE_K = 0x2752,   // molar entropy joule per mole kelvin
            ATT_UUID_EXPOSURE_C_PER_KG = 0x2753,   // exposure coulomb per kilogram
            ATT_UUID_DOSE_RATE_GRAY_PER_SEC = 0x2754,   // absorbed dose rate gray per second
            ATT_UUID_RT_INTENSITY_W_PER_STER = 0x2755,   // radiant intensity watt per steradian
            ATT_UUID_RCE_W_PER_SQ_METER_STER = 0x2756,   // radiance watt per square meter steradian
            ATT_UUID_CATALYTIC_KATAL_PER_CU_M = 0x2757,   // catalytic activity concentration katal per cubic metre
            ATT_UUID_TIME_MIN = 0x2760,   // time minute
            ATT_UUID_TIME_HR = 0x2761,   // time hour
            ATT_UUID_TIME_DAY = 0x2762,   // time day
            ATT_UUID_PLANE_ANGLE_DEG = 0x2763,   // plane angle degree
            ATT_UUID_PLANE_ANGLE_MIN = 0x2764,   // plane angle minute
            ATT_UUID_PLANE_ANGLE_SEC = 0x2765,   // plane angle second
            ATT_UUID_AREA_HECTARE = 0x2766,   // area hectare
            ATT_UUID_VOLUME_L = 0x2767,   // volume litre
            ATT_UUID_MASS_TONNE = 0x2768,   // mass tonne
            ATT_UUID_PRESSURE_BAR = 0x2780,   // pressure bar
            ATT_UUID_PRESSURE_MM = 0x2781,   // pressure millimetre of mercury
            ATT_UUID_LENGTH_ANGSTROM = 0x2782,   // length angstrom
            ATT_UUID_LENGTH_NAUTICAL_MILE = 0x2783,   // length nautical mile
            ATT_UUID_AREA_BARN = 0x2784,   // area barn
            ATT_UUID_VELOCITY_KNOT = 0x2785,   // velocity knot
            ATT_UUID_LOG_RADIO_QUANT_NEPER = 0x2786,   // logarithmic radio quantity neper
            ATT_UUID_LOG_RADIO_QUANT_BEL = 0x2787,   // logarithmic radio quantity bel
            ATT_UUID_LOG_RADIO_QUANT_DB = 0x2788,   // logarithmic radio quantity decibel
            ATT_UUID_LENGTH_YARD = 0x27A0,   // length yard
            ATT_UUID_LENGTH_PARSEC = 0x27A1,   // length parsec
            ATT_UUID_LENGTH_IN = 0x27A2,   // length inch
            ATT_UUID_LENGTH_FOOT = 0x27A3,   // length foot
            ATT_UUID_LENGTH_MILE = 0x27A4,   // length mile
            ATT_UUID_PRESSURE_POUND_PER_SQ_IN = 0x27A5,   // pressure pound-force per square inch
            ATT_UUID_VELOCITY_KPH = 0x27A6,   // velocity kilometre per hour
            ATT_UUID_VELOCITY_MPH = 0x27A7,   // velocity mile per hour
            ATT_UUID_ANG_VELOCITY_RPM = 0x27A8,   // angular velocity revolution per minute
            ATT_UUID_ENERGY_GRAM_CALORIE = 0x27A9,   // energy gram calorie
            ATT_UUID_ENERGY_KG_CALORIE = 0x27AA,   // energy kilogram calorie
            ATT_UUID_ENERGY_KILOWATT_HR = 0x27AB,   // energy kilowatt hour
            ATT_UUID_THERM_TEMP_F = 0x27AC,   // thermodynamic temperature degree Fahrenheit
            ATT_UUID_PERCENTAGE = 0x27AD,   // percentage
            ATT_UUID_PER_MILLE = 0x27AE,   // per mille
            ATT_UUID_PERIOD_BEATS_PER_MIN = 0x27AF,   // period beats per minute
            ATT_UUID_ELECTRIC_CHG_AMP_HRS = 0x27B0,   // electric charge ampere hours
            ATT_UUID_MASS_DENSITY_MG_PER_DL = 0x27B1,   // mass density milligram per decilitre
            ATT_UUID_MASS_DENSITY_MMOLE_PER_L = 0x27B2,   // mass density millimole per litre
            ATT_UUID_TIME_YEAR = 0x27B3,   // time year
            ATT_UUID_TIME_MONTH = 0x27B4,   // time month
        }
    }
}
