﻿using CySmart.Common.Base.Compatibility.Net20;
using CySmart.DongleCommunicator.API;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO.Ports;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;


// Cypress neglected to define Action<T> (with a single argument), so here goes...
namespace CySmart.Common.Base.Compatibility.Net20
{
    public delegate void Action<T1>(T1 arg1);
}

namespace Host
{
    public partial class mainWindow
    {
        ////////////////////////////////////////////////////////////////////////
        // This will give us the ICySmartDongleCommunicator instance
        private CySmartDongleMgr dongleMgr;

        ////////////////////////////////////////////////////////////////////////
        // This will give us the ICyBleMgr instance
        private ICySmartDongleCommunicator communicator;

        ////////////////////////////////////////////////////////////////////////
        // The BLE manager, which contains the ICyBleDeviceList, ICyBleSecurityMgr,
        //  and ICyBleDeviceAddressMgr instances
        public ICyBleMgr bleMgr
        {
            get { return communicator.BleMgr; }
        }
        // BLE manager callback - installed via bleMgr.RegisterBleMgrCallback()
        public bleMgrCallback bleMgrCb;
        // Scan callback - installed via bleMgr.StartScan()
        public scanCallback scanCb;

        ////////////////////////////////////////////////////////////////////////
        // The device list, which includes the whitelist, the bond list, and the resolving list
        public ICyBleDeviceList devList
        {
            get { return communicator.BleMgr.DeviceList; }
        }
        // Device list callback - installed via devList.RegisterDeviceListCallback()
        public deviceListCallback devListCb;

        ////////////////////////////////////////////////////////////////////////
        // The security manager
        public ICyBleSecurityMgr secMgr
        {
            get { return communicator.BleMgr.SecurityMgr; }
        }
        // Security manager callback - installed via secMgr.RegisterSecurityMgrCallback()
        public securityMgrCallback secMgrCb;

        ////////////////////////////////////////////////////////////////////////
        // The device address manager
        public ICyBleDeviceAddressMgr devAddrMgr
        {
            get { return communicator.BleMgr.DeviceAddressMgr; }
        }
        // Device address manager callback - installed via devAddrMgr.RegisterDeviceAddressMgrCallback()
        public deviceAddressMgrCallback devAddrMgrCb;

        ////////////////////////////////////////////////////////////////////////
        // The currently connected remote device, which contains the ICyGattClient
        //  and ICyL2CapMgr instances
        public ICyBleDevice connectedDevice;
        // Connected device callback - installed via connectedDevice.RegisterCallback()
        public deviceCallback devCb;

        ////////////////////////////////////////////////////////////////////////
        // The GATT client
        public ICyGattClient gattClient
        {
            get { return connectedDevice.GattClient; }
        }
        // GATT client callback - installed via gattClient.RegisterCallback()
        public gattClientCallback gattCb;
        // Primary service discovery callback - installed via gattClient.DiscoverPrimaryServices()
        //  and/or gattClient.DiscoverPrimaryServicesByUUID()
        public discoverPrimaryServiceCallback discPrimCb;
        // Included services discovery callback - installed via gattClient.FindIncludedServices()
        public findIncludedServicesCallback findIncCb;
        // Characteristics discovery callback - installed via gattClient.DiscoverCharacteristics()
        //  and/or gattClient.DiscoverCharacteristicsByUUID()
        public discoverCharacteristicsCallback discCharCb;
        // Descriptor discovery callback - installed via gattClient.DiscoverDescriptors()
        public discoverDescriptorsCallback discDescCb;

        ////////////////////////////////////////////////////////////////////////
        // The L2CAP manager
        public ICyL2CapMgr l2capMgr
        {
            get { return connectedDevice.L2CapMgr; }
        }
        // L2CAP manager callback - installed via l2capMgr.RegisterL2CapCallback()
        public l2capMgrCallback l2capMgrCb;

        public CyScanRecord device;
        public CyScanRecord highlightedDevice;
        public CyDiscoverAllServicesResult gattData;
    }


    ////////////////////////////////////////////////////////////////////////////
    // BLE manager callback class
    ////////////////////////////////////////////////////////////////////////////
    public class bleMgrCallback : CyBleMgrCallback
    {
        public CySmart.Common.Base.Compatibility.Net20.Action<CyStatus> SetDeviceIoCapabilitiesHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyBleDeviceIoCapabilities, CyStatus> GetDeviceIoCapabilitiesHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<sbyte, CyStatus> GetRssiHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyConnectResult, CyStatus> ConnectedHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyStatus> CancelledHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyBleBdAddress, CyStatus> DisconnectedHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyRegisteredPsm, CyStatus> RegisterPsmHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyStatus> UnregisterPsmHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyStatus> SetHostChannelClassificationHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyStatus> SetTxPowerHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyTxPowerInfo, CyStatus> GetTxPowerHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyDefaultDataLengthResult, CyStatus> GetDefaultDataLengthHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyStatus> SetSuggestedDataLengthHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyConvertOctetToTimeResult, CyStatus> ConvertDataLengthOctetToTimeHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyStatus> SetResolvableAddressTimeoutHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyStatus> SetAddressResolutionControlHandler { get; set; }

        // Reports the status of ICyBleMgr.SetDeviceIoCapabilities()
        public override void OnSetDeviceIoCapabilities(CyStatus status)
        {
            if (SetDeviceIoCapabilitiesHandler != null)
            {
                SetDeviceIoCapabilitiesHandler(status);
            }
        }

        // Reports the IO capabilities of the local device
        public override void OnGetDeviceIoCapabilities(CyBleDeviceIoCapabilities ioCapabilities, CyStatus status)
        {
            if (GetDeviceIoCapabilitiesHandler != null)
            {
                GetDeviceIoCapabilitiesHandler(ioCapabilities, status);
            }
        }

        // Reports the RSSI and the status of ICyBleMgr.GetRSSI()
        public override void OnGetRssi(sbyte rssi, CyStatus status)
        {
            if (GetRssiHandler != null)
            {
                GetRssiHandler(rssi, status);
            }
        }

        // Reports the result and status of ICyBleMgr.Connect()
        public override void OnConnected(CyConnectResult result, CyStatus status)
        {
            if (ConnectedHandler != null)
            {
                ConnectedHandler(result, status);
            }
        }

        // Reports the status of ICyBleMgr.CancelConnection()
        public override void OnCancelled(CyStatus status)
        {
            if (CancelledHandler != null)
            {
                CancelledHandler(status);
            }
        }

        // Reports the status of ICyBleMgr.Disconnect()
        public override void OnDisconnected(CyBleBdAddress deviceAddress, CyStatus status)
        {
            if (DisconnectedHandler != null)
            {
                DisconnectedHandler(deviceAddress, status);
            }
        }

        // Reports the registered PSM
        public override void OnRegisterPsm(CyRegisteredPsm psm, CyStatus status)
        {
            if (RegisterPsmHandler != null)
            {
                RegisterPsmHandler(psm, status);
            }
        }

        // Reports the status of ICyBleMgr.UnregisterPsm()
        public override void OnUnregisterPsm(CyStatus status)
        {
            if (UnregisterPsmHandler != null)
            {
                UnregisterPsmHandler(status);
            }
        }

        // Reports the status of ICyBleMgr.SetHostChannelClassification()
        public override void OnSetHostChannelClassification(CyStatus status)
        {
            if (SetHostChannelClassificationHandler != null)
            {
                SetHostChannelClassificationHandler(status);
            }
        }

        // Reports the status of ICyBleMgr.SetTxPower()
        public override void OnSetTxPower(CyStatus status)
        {
            if (SetTxPowerHandler != null)
            {
                SetTxPowerHandler(status);
            }
        }

        // Reports the channel transmission power
        public override void OnGetTxPower(CyTxPowerInfo result, CyStatus status)
        {
            if (GetTxPowerHandler != null)
            {
                GetTxPowerHandler(result, status);
            }
        }

        // Reports the default data length
        public override void OnGetDefaultDataLength(CyDefaultDataLengthResult result, CyStatus status)
        {
            if (GetDefaultDataLengthHandler != null)
            {
                GetDefaultDataLengthHandler(result, status);
            }
        }

        // Reports the status of ICyBleMgr.SetSuggestedDataLength()
        public override void OnSetSuggestedDataLength(CyStatus status)
        {
            if (SetSuggestedDataLengthHandler != null)
            {
                SetSuggestedDataLengthHandler(status);
            }
        }

        // Reports the result of the octet-to-time conversion
        public override void OnConvertDataLengthOctetToTime(CyConvertOctetToTimeResult result, CyStatus status)
        {
            if (ConvertDataLengthOctetToTimeHandler != null)
            {
                ConvertDataLengthOctetToTimeHandler(result, status);
            }
        }

        // Reports the status of ICyBleMgr.SetResolvableAddressTimeout()
        public override void OnSetResolvableAddressTimeout(CyStatus status)
        {
            if (SetResolvableAddressTimeoutHandler != null)
            {
                SetResolvableAddressTimeoutHandler(status);
            }
        }

        // Reports the status of ICyBleMgr.SetAddressResolutionControl()
        public override void OnSetAddressResolutionControl(CyStatus status)
        {
            if (SetAddressResolutionControlHandler != null)
            {
                SetAddressResolutionControlHandler(status);
            }
        }
    }

    ////////////////////////////////////////////////////////////////////////////
    // Scan callback class
    ////////////////////////////////////////////////////////////////////////////
    public class scanCallback : CyScanCallback
    {
        public CySmart.Common.Base.Compatibility.Net20.Action<CyScanResult> ScanResultHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyScanStatus> ScanStatusChangedHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyStatus> StartScanErrorHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyStatus> StopScanErrorHandler { get; set; }

        // Reports the discovered BLE devices
        public override void OnScanResult(CyScanResult result)
        {
            if (ScanResultHandler != null)
            {
                ScanResultHandler(result);
            }
        }

        // Reports the scan status
        public override void OnScanStatusChanged(CyScanStatus scanStatus)
        {
            if (ScanStatusChangedHandler != null)
            {
                ScanStatusChangedHandler(scanStatus);
            }
        }

        // Reports the IcyBleMgr.StartScan() error; not called when there is no error.
        public override void OnStartScanError(CyStatus status)
        {
            if (StartScanErrorHandler != null)
            {
                StartScanErrorHandler(status);
            }
        }

        // Reports the IcyBleMgr.StopScan() error; not called when there is no error.
        public override void OnStopScanError(CyStatus status)
        {
            if (StopScanErrorHandler != null)
            {
                StopScanErrorHandler(status);
            }
        }
    }

    ////////////////////////////////////////////////////////////////////////////
    // Device list callback class
    ////////////////////////////////////////////////////////////////////////////
    public class deviceListCallback : CyDeviceListCallback
    {
        public CySmart.Common.Base.Compatibility.Net20.Action<CyWhitelistDevice, CyStatus> AddDeviceToWhitelistHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyStatus> RemoveDeviceFromWhitelistHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyStatus> ClearWhitelistHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<List<CyWhitelistDevice>, CyStatus> GetWhitelistDevicesHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<List<CyBondListDevice>, CyStatus> GetBondListDevicesHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyResolvingListDevice, CyStatus> AddDeviceToResolvingListHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyStatus> RemoveDeviceFromResolvingListHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyStatus> ClearResolvingListHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<List<CyResolvingListDevice>, CyStatus> GetResolvingListDevicesHandler { get; set; }

        // Reports the device added to the whitelist
        public override void OnAddDeviceToWhitelist(CyWhitelistDevice device, CyStatus status)
        {
            if (AddDeviceToWhitelistHandler != null)
            {
                AddDeviceToWhitelistHandler(device, status);
            }
        }

        // Reports the status of ICyBleDeviceList.RemoveDeviceFromWhitelist()
        public override void OnRemoveDeviceFromWhitelist(CyStatus status)
        {
            if (RemoveDeviceFromWhitelistHandler != null)
            {
                RemoveDeviceFromWhitelistHandler(status);
            }
        }

        // Reports the status of ICyBleDeviceList.ClearWhitelist()
        public override void OnClearWhitelist(CyStatus status)
        {
            if (ClearWhitelistHandler != null)
            {
                ClearWhitelistHandler(status);
            }
        }

        // Reports the whitelist devices
        public override void OnGetWhitelistDevices(List<CyWhitelistDevice> whitelist, CyStatus status)
        {
            if (GetWhitelistDevicesHandler != null)
            {
                GetWhitelistDevicesHandler(whitelist, status);
            }
        }

        // Reports the bond list devices
        public override void OnGetBondListDevices(List<CyBondListDevice> bondList, CyStatus status)
        {
            if (GetBondListDevicesHandler != null)
            {
                GetBondListDevicesHandler(bondList, status);
            }
        }

        // Reports the device added to the resolving list
        public override void OnAddDeviceToResolvingList(CyResolvingListDevice device, CyStatus status)
        {
            if (AddDeviceToResolvingListHandler != null)
            {
                AddDeviceToResolvingListHandler(device, status);
            }
        }

        // Reports the status of ICyBleDeviceList.RemoveDeviceFromResolvinglist()
        public override void OnRemoveDeviceFromResolvingList(CyStatus status)
        {
            if (RemoveDeviceFromResolvingListHandler != null)
            {
                RemoveDeviceFromResolvingListHandler(status);
            }
        }

        // Reports the status of ICyBleDeviceList.ClearResolvinglist()
        public override void OnClearResolvingList(CyStatus status)
        {
            if (ClearResolvingListHandler != null)
            {
                ClearResolvingListHandler(status);
            }
        }

        // Reports the resolving list devices
        public override void OnGetResolvingListDevices(List<CyResolvingListDevice> resolvingList, CyStatus status)
        {
            if (GetResolvingListDevicesHandler != null)
            {
                GetResolvingListDevicesHandler(resolvingList, status);
            }
        }
    }

    ////////////////////////////////////////////////////////////////////////////
    // Security manager callback class
    ////////////////////////////////////////////////////////////////////////////
    public class securityMgrCallback : CySecurityMgrCallback
    {
        public CySmart.Common.Base.Compatibility.Net20.Action<CyAuthenticationKeys, CyStatus> GetAuthenticationKeysHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyAuthenticationKeys, CyStatus> GenerateAuthenticationKeysHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyStatus> SetAuthenticationKeysHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyGetPeerDeviceAuthenticationKeyResult, CyStatus> GetPeerDeviceAuthenticationKeysHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyResolveAddressResult, CyStatus> ResolvePeerDeviceAddressHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CySecureConnectionOobDataResult, CyStatus> GenerateSecureConnectionOobDataHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyStatus> GenerateLocalECDHKeyHandler { get; set; }

        // Reports the local authentication keys
        public override void OnGetAuthenticationKeys(CyAuthenticationKeys result, CyStatus status)
        {
            if (GetAuthenticationKeysHandler != null)
            {
                GetAuthenticationKeysHandler(result, status);
            }
        }

        // Reports the generated authentication keys
        public override void OnGenerateAuthenticationKeys(CyAuthenticationKeys result, CyStatus status)
        {
            if (GenerateAuthenticationKeysHandler != null)
            {
                GenerateAuthenticationKeysHandler(result, status);
            }
        }

        // Reports the status of ICyBleSecurityMgr.SetAuthenticationKeys()
        public override void OnSetAuthenticationKeys(CyStatus status)
        {
            if (SetAuthenticationKeysHandler != null)
            {
                SetAuthenticationKeysHandler(status);
            }
        }

        // Reports the peer device authentication keys
        public override void OnGetPeerDeviceAuthenticationKeys(CyGetPeerDeviceAuthenticationKeyResult result, CyStatus status)
        {
            if (GetPeerDeviceAuthenticationKeysHandler != null)
            {
                GetPeerDeviceAuthenticationKeysHandler(result, status);
            }
        }

        // Reports the status of ICyBleSecurityMgr.ResolvePeerDeviceAddress()
        public override void OnResolvePeerDeviceAddress(CyResolveAddressResult result, CyStatus status)
        {
            if (ResolvePeerDeviceAddressHandler != null)
            {
                ResolvePeerDeviceAddressHandler(result, status);
            }
        }

        // Reports the generated secure connection OOB data
        public override void OnGenerateSecureConnectionOobData(CySecureConnectionOobDataResult result, CyStatus status)
        {
            if (GenerateSecureConnectionOobDataHandler != null)
            {
                GenerateSecureConnectionOobDataHandler(result, status);
            }
        }

        // Reports the status of ICyBleSecurityMgr.GenerateLocalECDHKey()
        public override void OnGenerateLocalECDHKey(CyStatus status)
        {
            if (GenerateLocalECDHKeyHandler != null)
            {
                GenerateLocalECDHKeyHandler(status);
            }
        }
    }

    ////////////////////////////////////////////////////////////////////////////
    // Device address manager callback class
    ////////////////////////////////////////////////////////////////////////////
    public class deviceAddressMgrCallback : CyDeviceAddressMgrCallback
    {
        public CySmart.Common.Base.Compatibility.Net20.Action<CyGenerateBdAddressResult, CyStatus> BdAddressGeneratedHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyBleBdAddress, CyStatus> GetBdAddressHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyStatus> SetBdAddressHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyStatus> SetIdAddressHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyResolvableAddressResult, CyStatus> GetPeerResolvableAddressHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyResolvableAddressResult, CyStatus> GetLocalResolvableAddressHandler { get; set; }

        // Reports the generated Bluetooth device address
        public override void OnBdAddressGenerated(CyGenerateBdAddressResult result, CyStatus status)
        {
            if (BdAddressGeneratedHandler != null)
            {
                BdAddressGeneratedHandler(result, status);
            }
        }

        // Reports the Bluetooth device address of the local device (dongle)
        public override void OnGetBdAddress(CyBleBdAddress address, CyStatus status)
        {
            if (GetBdAddressHandler != null)
            {
                GetBdAddressHandler(address, status);
            }
        }

        // Reports the status of ICyBleDeviceAddressMgr.SetBdAddress()
        public override void OnSetBdAddress(CyStatus status)
        {
            if (SetBdAddressHandler != null)
            {
                SetBdAddressHandler(status);
            }
        }

        // Reports the status of ICyBleDeviceAddressMgr.SetIdAddress()
        public override void OnSetIdAddress(CyStatus status)
        {
            if (SetIdAddressHandler != null)
            {
                SetIdAddressHandler(status);
            }
        }

        // Reports the resolvable address of the peer device
        public override void OnGetPeerResolvableAddress(CyResolvableAddressResult result, CyStatus status)
        {
            if (GetPeerResolvableAddressHandler != null)
            {
                GetPeerResolvableAddressHandler(result, status);
            }
        }

        // Reports the resolvable address of the local device
        public override void OnGetLocalResolvableAddress(CyResolvableAddressResult result, CyStatus status)
        {
            if (GetLocalResolvableAddressHandler != null)
            {
                GetLocalResolvableAddressHandler(result, status);
            }
        }
    }

    ////////////////////////////////////////////////////////////////////////////
    // BLE device callback class
    ////////////////////////////////////////////////////////////////////////////
    public class deviceCallback : CyBleDeviceCallback
    {
        public CySmart.Common.Base.Compatibility.Net20.Action<CyConnectionParameters> UpdateConnectionParameterRequestReceivedHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyStatus> UpdateConnectionParameterHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyCurrentConnectionParameters> ConnectionParameterChangedHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyStatus> SetDataLengthHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyCurrentDataLength> DataLengthChangedHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyPairSettings> PairRequestReceivedHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyPairSettings> PairingSettingsNegotiatedHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyPasskeyEntryResponse> PasskeyEntryRequestHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyPasskeyDisplayInfo> PasskeyDisplayRequestHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyNumericComparisonResponse> NumericComparisonRequestHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyStatus> SendKeyPressNotificationHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyStatus> PairingCompletedHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyStatus> SetOobHandler { get; set; }

        // Connection parameter update request received from remote device
        public override void OnUpdateConnectionParameterRequestReceived(CyConnectionParameters request)
        {
            if (UpdateConnectionParameterRequestReceivedHandler != null)
            {
                UpdateConnectionParameterRequestReceivedHandler(request);
            }
        }

        // Reports the status of ICyBleDevice.UpdateConnectionParameter()
        public override void OnUpdateConnectionParameter(CyStatus status)
        {
            if (UpdateConnectionParameterHandler != null)
            {
                UpdateConnectionParameterHandler(status);
            }
        }

        // Reports the updated connection parameter values
        public override void OnConnectionParameterChanged(CyCurrentConnectionParameters connectionParameters)
        {
            if (ConnectionParameterChangedHandler != null)
            {
                ConnectionParameterChangedHandler(connectionParameters);
            }
        }

        // Reports the status of ICyBleDevice.SetDataLength()
        public override void OnSetDataLength(CyStatus status)
        {
            if (SetDataLengthHandler != null)
            {
                SetDataLengthHandler(status);
            }
        }

        // Reports the updated data length for the connection
        public override void OnDataLengthChanged(CyCurrentDataLength dataLength)
        {
            if (DataLengthChangedHandler != null)
            {
                DataLengthChangedHandler(dataLength);
            }
        }

        // Reports the pairing request received from the remote device
        public override void OnPairRequestReceived(CyPairSettings settings)
        {
            if (PairRequestReceivedHandler != null)
            {
                PairRequestReceivedHandler(settings);
            }
        }

        // Reports the negotiated pairing settings
        public override void OnPairingSettingsNegotiated(CyPairSettings settings)
        {
            if (PairingSettingsNegotiatedHandler != null)
            {
                PairingSettingsNegotiatedHandler(settings);
            }
        }

        // Passkey entry request
        public override void OnPasskeyEntryRequest(CyPasskeyEntryResponse response)
        {
            if (PasskeyEntryRequestHandler != null)
            {
                PasskeyEntryRequestHandler(response);
            }
        }

        // Passkey display request
        public override void OnPasskeyDisplayRequest(CyPasskeyDisplayInfo info)
        {
            if (PasskeyDisplayRequestHandler != null)
            {
                PasskeyDisplayRequestHandler(info);
            }
        }

        // Secure connection numeric comparison request
        public override void OnNumericComparisonRequest(CyNumericComparisonResponse response)
        {
            if (NumericComparisonRequestHandler != null)
            {
                NumericComparisonRequestHandler(response);
            }
        }

        // Reports the status of ICyBleDevice.SendKeyPressNotification()
        public override void OnSendKeyPressNotification(CyStatus status)
        {
            if (SendKeyPressNotificationHandler != null)
            {
                SendKeyPressNotificationHandler(status);
            }
        }

        // Reports the status of ICyBleDevice.Pair()
        public override void OnPairingCompleted(CyStatus status)
        {
            if (PairingCompletedHandler != null)
            {
                PairingCompletedHandler(status);
            }
        }

        // Reports the status of ICyBleDevice.SetOob()
        public override void OnSetOob(CyStatus status)
        {
            if (SetOobHandler != null)
            {
                SetOobHandler(status);
            }
        }
    }

    ////////////////////////////////////////////////////////////////////////////
    // GATT client callback class
    ////////////////////////////////////////////////////////////////////////////
    public class gattClientCallback : CyGattClientCallback
    {
        public CySmart.Common.Base.Compatibility.Net20.Action<CyGattExchangeMtuResult, CyStatus> GattMtuExchangedHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyDiscoverAllServicesResult, CyStatus> ServiceDiscoveredHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyCharacteristicChangedInfo> CharacteristicChangedHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyGattReadResult, CyStatus> CharacteristicReadHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyReadCharacteristicByUUIDResult, CyStatus> CharacteristicReadByUUIDHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyReadMultipleCharacteristicResult, CyStatus> ReadMultipleCharacteristicsHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyGattWriteResult, CyStatus> CharacteristicWriteHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyStatus> ReliableWriteCompletedHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyWriteBufferFullResponse> WriteBufferFullHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyGattReadResult, CyStatus> DescriptorReadHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyGattWriteResult, CyStatus> DescriptorWriteHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyStatus> GattStopHandler { get; set; }

        // Reports the negotiated GATT MTU size
        public override void OnGattMtuExchanged(CyGattExchangeMtuResult result, CyStatus status)
        {
            if (GattMtuExchangedHandler != null)
            {
                GattMtuExchangedHandler(result, status);
            }
        }

        // Reports all the discovered services
        public override void OnServiceDiscovered(CyDiscoverAllServicesResult result, CyStatus status)
        {
            if (ServiceDiscoveredHandler != null)
            {
                ServiceDiscoveredHandler(result, status);
            }
        }

        // Reports the received characteristic value notifications and indications
        public override void OnCharacteristicChanged(CyCharacteristicChangedInfo info)
        {
            if (CharacteristicChangedHandler != null)
            {
                CharacteristicChangedHandler(info);
            }
        }

        // Reports the characteristic value that was read
        public override void OnCharacteristicRead(CyGattReadResult result, CyStatus status)
        {
            if (CharacteristicReadHandler != null)
            {
                CharacteristicReadHandler(result, status);
            }
        }

        // Reports the value of characteristics with a specific UUID
        public override void OnCharacteristicReadByUUID(CyReadCharacteristicByUUIDResult result, CyStatus status)
        {
            if (CharacteristicReadByUUIDHandler != null)
            {
                CharacteristicReadByUUIDHandler(result, status);
            }
        }

        // Reports the value read by ICyGattClient.ReadMultipleCharacteristic()
        public override void OnReadMultipleCharacteristics(CyReadMultipleCharacteristicResult result, CyStatus status)
        {
            if (ReadMultipleCharacteristicsHandler != null)
            {
                ReadMultipleCharacteristicsHandler(result, status);
            }
        }

        // Reports the status of a characteristic value write
        public override void OnCharacteristicWrite(CyGattWriteResult result, CyStatus status)
        {
            if (CharacteristicWriteHandler != null)
            {
                CharacteristicWriteHandler(result, status);
            }
        }

        // Reports the result of a reliable write
        public override void OnReliableWriteCompleted(CyStatus status)
        {
            if (ReliableWriteCompletedHandler != null)
            {
                ReliableWriteCompletedHandler(status);
            }
        }

        // Invoked when the peer device reports that the write buffer is full
        public override void OnWriteBufferFull(CyWriteBufferFullResponse response)
        {
            if (WriteBufferFullHandler != null)
            {
                WriteBufferFullHandler(response);
            }
        }

        // Reports the descriptor value that was read
        public override void OnDescriptorRead(CyGattReadResult result, CyStatus status)
        {
            if (DescriptorReadHandler != null)
            {
                DescriptorReadHandler(result, status);
            }
        }

        // Reports the status of a descriptor value write
        public override void OnDescriptorWrite(CyGattWriteResult result, CyStatus status)
        {
            if (DescriptorWriteHandler != null)
            {
                DescriptorWriteHandler(result, status);
            }
        }

        // Reports the status of ICyGattClient.GattStop()
        public override void OnGattStop(CyStatus status)
        {
            if (GattStopHandler != null)
            {
                GattStopHandler(status);
            }
        }
    }

    ////////////////////////////////////////////////////////////////////////////
    // Discover primary services callback class
    ////////////////////////////////////////////////////////////////////////////
    public class discoverPrimaryServiceCallback : CyDiscoverPrimaryServiceCallback
    {
        public CySmart.Common.Base.Compatibility.Net20.Action<List<CyGattService>, CyStatus> PrimaryServiceDiscoveredHandler { get; set; }

        // Reports the discovered primary services
        public override void OnPrimaryServiceDiscovered(List<CyGattService> services, CyStatus status)
        {
            if (PrimaryServiceDiscoveredHandler != null)
            {
                PrimaryServiceDiscoveredHandler(services, status);
            }
        }
    }

    ////////////////////////////////////////////////////////////////////////////
    // Find included services callback class
    ////////////////////////////////////////////////////////////////////////////
    public class findIncludedServicesCallback : CyFindIncludedServicesCallback
    {
        public CySmart.Common.Base.Compatibility.Net20.Action<CyFindIncludedServicesResult, CyStatus> FindIncludedServicesHandler { get; set; }

        // Reports the discovered included services
        public override void OnFindIncludedServices(CyFindIncludedServicesResult result, CyStatus status)
        {
            if (FindIncludedServicesHandler != null)
            {
                FindIncludedServicesHandler(result, status);
            }
        }
    }

    ////////////////////////////////////////////////////////////////////////////
    // Discover characteristics callback class
    ////////////////////////////////////////////////////////////////////////////
    public class discoverCharacteristicsCallback : CyDiscoverCharacteristicsCallback
    {
        public CySmart.Common.Base.Compatibility.Net20.Action<CyDiscoverCharacteristicsResult, CyStatus> CharacteristicsDiscoveredHandler { get; set; }

        // Reports the discovered characteristics
        public override void OnCharacteristicsDiscovered(CyDiscoverCharacteristicsResult result, CyStatus status)
        {
            if (CharacteristicsDiscoveredHandler != null)
            {
                CharacteristicsDiscoveredHandler(result, status);
            }
        }
    }

    ////////////////////////////////////////////////////////////////////////////
    // Discover descriptors callback class
    ////////////////////////////////////////////////////////////////////////////
    public class discoverDescriptorsCallback : CyDiscoverDescriptorsCallback
    {
        public CySmart.Common.Base.Compatibility.Net20.Action<CyDiscoverDescriptorsResult, CyStatus> DescriptorDiscoveredHandler { get; set; }

        // Reports the discovered descriptors
        public override void OnDescriptorDiscovered(CyDiscoverDescriptorsResult result, CyStatus status)
        {
            if (DescriptorDiscoveredHandler != null)
            {
                DescriptorDiscoveredHandler(result, status);
            }
        }
    }

    ////////////////////////////////////////////////////////////////////////////
    // L2CAP manager callback class
    ////////////////////////////////////////////////////////////////////////////
    public class l2capMgrCallback : CyL2CapMgrCallback
    {
        public CySmart.Common.Base.Compatibility.Net20.Action<CyL2CapConnectionResponseInfo> ChannelConnectionIndicationHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyL2CapDisconnectIndicationInfo> ChannelDisconnectIndicationHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyL2CapDisconnectConfirmation, CyStatus> DisconnectChannelHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<ICyL2CapChannel, CyStatus> ChannelEstablishedHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyL2CapDataReceivedInfo> DataReceivedHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyL2CapReceiveCreditLowInfo> ReceiveCreditLowIndicationHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<CyL2CapTransmitCreditInfo> TransmitCreditIndicationHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<ushort, CyStatus> SendDataHandler { get; set; }
        public CySmart.Common.Base.Compatibility.Net20.Action<ushort, CyStatus> SendCreditsHandler { get; set; }

        // Reports an L2CAP channel connection request received from a peer device
        public override void OnChannelConnectionIndication(CyL2CapConnectionResponseInfo response)
        {
            if (ChannelConnectionIndicationHandler != null)
            {
                ChannelConnectionIndicationHandler(response);
            }
        }

        // Reports an L2CAP channel disconnect initiated by the local device stack or by the peer device
        public override void OnChannelDisconnectIndication(CyL2CapDisconnectIndicationInfo info)
        {
            if (ChannelDisconnectIndicationHandler != null)
            {
                ChannelDisconnectIndicationHandler(info);
            }
        }

        // Reports the status of ICyL2CapMgr.DisconnectChannel()
        public override void OnDisconnectChannel(CyL2CapDisconnectConfirmation result, CyStatus status)
        {
            if (DisconnectChannelHandler != null)
            {
                DisconnectChannelHandler(result, status);
            }
        }

        // Reports the established L2CAP channel
        public override void OnChannelEstablished(ICyL2CapChannel channel, CyStatus status)
        {
            if (ChannelEstablishedHandler != null)
            {
                ChannelEstablishedHandler(channel, status);
            }
        }

        // Reports the data received over an L2CAP channel
        public override void OnDataReceived(CyL2CapDataReceivedInfo info)
        {
            if (DataReceivedHandler != null)
            {
                DataReceivedHandler(info);
            }
        }

        // Reports that the receive credit of an L2CAP channel is low
        public override void OnReceiveCreditLowIndication(CyL2CapReceiveCreditLowInfo info)
        {
            if (ReceiveCreditLowIndicationHandler != null)
            {
                ReceiveCreditLowIndicationHandler(info);
            }
        }

        // Reports the transmit flow credit received from a peer device, for an L2CAP channel
        public override void OnTransmitCreditIndication(CyL2CapTransmitCreditInfo info)
        {
            if (TransmitCreditIndicationHandler != null)
            {
                TransmitCreditIndicationHandler(info);
            }
        }

        // Reports the status of ICyL2CapMgr.SendData()
        public override void OnSendData(ushort channelID, CyStatus status)
        {
            if (SendDataHandler != null)
            {
                SendDataHandler(channelID, status);
            }
        }

        // Reports the status of ICyL2CapMgr.SendCredits()
        public override void OnSendCredits(ushort channelID, CyStatus status)
        {
            if (SendCreditsHandler != null)
            {
                SendCreditsHandler(channelID, status);
            }
        }
    }
}
