﻿#undef debugTestGatt
#define debugTestGatt

using CySmart.Common.Base.Compatibility.Net20;
using CySmart.DongleCommunicator.API;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO.Ports;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;


namespace Host
{
    public static class Gatt
    {
        private static Form gattBox;

        public static void gattDialogBox()
        {
            gattBox = new Form();
            gattBox.Name = "gattBox";
            gattBox.Width = 600;
            gattBox.Height = 500;
            gattBox.Text = "Remote GATT Database";
            gattBox.Icon = Program.mf.Icon;
            gattBox.FormClosed += (sender, e) =>
            {
                Utilities.SetState(Constants.hostStates.HOST_STATE_DEVICE_CONNECTED);
                gattBox.Dispose();
            };

            TreeView gattTree = new TreeView()
            {
                Dock = System.Windows.Forms.DockStyle.Fill,
            };

            Utilities.SetState(Constants.hostStates.HOST_STATE_GATT_SHOWN);

            gattTree.BeginUpdate();
            int svcIndex = 0;
            foreach (var svc in Program.mf.gattData.Services)
            {
                gattTree.Nodes.Add(GetServiceName(svc.UUID));
                gattTree.Nodes[svcIndex].Nodes.Add(string.Format("Primary? {0}", svc.IsPrimaryService));
                gattTree.Nodes[svcIndex].Nodes.Add(string.Format("Start Handle: 0x{0:X4}", svc.StartHandle));
                gattTree.Nodes[svcIndex].Nodes.Add(string.Format("End Handle: 0x{0:X4}", svc.EndHandle));
                gattTree.Nodes[svcIndex].Nodes.Add("Characteristics:");
                int chrIndex = 0;
                foreach (var chr in svc.Characteristics)
                {
                    gattTree.Nodes[svcIndex].Nodes[3].Nodes.Add(GetCharacteristicName(chr.UUID));
                    gattTree.Nodes[svcIndex].Nodes[3].Nodes[chrIndex].Nodes.Add(string.Format("Declaration Handle: 0x{0:X4}", chr.DeclarationHandle));
                    gattTree.Nodes[svcIndex].Nodes[3].Nodes[chrIndex].Nodes.Add(string.Format("Value Handle: 0x{0:X4}", chr.Handle));
                    gattTree.Nodes[svcIndex].Nodes[3].Nodes[chrIndex].Nodes.Add(GetProperties(chr.Properties));
                    gattTree.Nodes[svcIndex].Nodes[3].Nodes[chrIndex].Nodes.Add("Included by " + GetServiceName(chr.Service.UUID));
                    if (chr.Descriptors.Count != 0)
                    {
                        gattTree.Nodes[svcIndex].Nodes[3].Nodes[chrIndex].Nodes.Add("Descriptors:");
                        int descIndex = 0;
                        foreach (var desc in chr.Descriptors)
                        {
                            gattTree.Nodes[svcIndex].Nodes[3].Nodes[chrIndex].Nodes[4].Nodes.Add(GetDescriptorName(desc.UUID));
                            gattTree.Nodes[svcIndex].Nodes[3].Nodes[chrIndex].Nodes[4].Nodes[descIndex].Nodes.Add(string.Format("Handle: 0x{0:X4}", desc.Handle));
                            gattTree.Nodes[svcIndex].Nodes[3].Nodes[chrIndex].Nodes[4].Nodes[descIndex].Nodes.Add(GetCharacteristicName(desc.Characteristic.UUID));
                            ++descIndex;
                        }
                    }
                    ++chrIndex;
                }
                if (svc.IncludedServices.Count != 0)
                {
                    gattTree.Nodes[svcIndex].Nodes.Add("Included Services:");
                    int incIndex = 0;
                    foreach (var inc in svc.IncludedServices)
                    {
                        gattTree.Nodes[svcIndex].Nodes[4].Nodes.Add(GetServiceName(inc.IncludedService.UUID));
                        gattTree.Nodes[svcIndex].Nodes[4].Nodes[incIndex].Nodes.Add("Included by " + GetServiceName(inc.Service.UUID));
                        gattTree.Nodes[svcIndex].Nodes[4].Nodes[incIndex].Nodes.Add(string.Format("Handle where included: 0x{0:X}", inc.IncludedHandle));
                        ++incIndex;
                    }
                }
                ++svcIndex;
            }
            gattTree.EndUpdate();

            gattBox.Controls.Add(gattTree);
            gattBox.ShowDialog();
        }

        private static string GetServiceName(CyUUID uuid)
        {
            string svcString;
            string name;

            if (uuid.IsUUID16Valid)
            {
                svcString = string.Format("UUID: 0x{0:X4} - ", uuid.UUID16);
                switch ((Constants.uuids)uuid.UUID16)
                {
                    case Constants.uuids.ATT_UUID_GAP_SERVICE:
                        name = "GAP";
                        break;

                    case Constants.uuids.ATT_UUID_GATT_SERVICE:
                        name = "GATT";
                        break;

                    case Constants.uuids.ATT_UUID_IMMEDIATE_ALERT_SERVICE:
                        name = "Immediate Alert";
                        break;

                    case Constants.uuids.ATT_UUID_LINK_LOSS_SERVICE:
                        name = "Link Loss";
                        break;

                    case Constants.uuids.ATT_UUID_TX_POWER_SERVICE:
                        name = "TX Power";
                        break;

                    case Constants.uuids.ATT_UUID_CURRENT_TIME_SERVICE:
                        name = "Current Time";
                        break;

                    case Constants.uuids.ATT_UUID_REF_TIME_UPDATE_SERVICE:
                        name = "Reference Time Update";
                        break;

                    case Constants.uuids.ATT_UUID_DST_CHANGE_SERVICE:
                        name = "Next DST Change";
                        break;

                    case Constants.uuids.ATT_UUID_GLUCOSE_SERVICE:
                        name = "Glucose";
                        break;

                    case Constants.uuids.ATT_UUID_HEALTH_THERM_SERVICE:
                        name = "Health Thermometer";
                        break;

                    case Constants.uuids.ATT_UUID_DEVICE_INFO_SERVICE:
                        name = "Device Information";
                        break;

                    case Constants.uuids.ATT_UUID_NETWORK_AVAIL_SERVICE:
                        name = "Network Availability";
                        break;

                    case Constants.uuids.ATT_UUID_WATCHDOG_SERVICE:
                        name = "Watchdog";
                        break;

                    case Constants.uuids.ATT_UUID_HEART_RATE_SERVICE:
                        name = "Heart Rate";
                        break;

                    case Constants.uuids.ATT_UUID_PHONE_ALERT_SERVICE:
                        name = "Phone Alert Status";
                        break;

                    case Constants.uuids.ATT_UUID_BATTERY_SERVICE:
                        name = "Battery";
                        break;

                    case Constants.uuids.ATT_UUID_BLOOD_PRESSURE_SERVICE:
                        name = "Blood Pressure";
                        break;

                    case Constants.uuids.ATT_UUID_ALERT_NOTIF_SERVICE:
                        name = "Alert Notification";
                        break;

                    case Constants.uuids.ATT_UUID_HID_SERVICE:
                        name = "HID";
                        break;

                    case Constants.uuids.ATT_UUID_SCAN_PARAM_SERVICE:
                        name = "Scan Parameter";
                        break;

                    case Constants.uuids.ATT_UUID_RUNNING_SPEED_SERVICE:
                        name = "Running Speed";
                        break;

                    case Constants.uuids.ATT_UUID_CYCLING_SPEED_SERVICE:
                        name = "Cycling Speed";
                        break;

                    case Constants.uuids.ATT_UUID_CYCLING_POWER_SERVICE:
                        name = "Cycling Power";
                        break;

                    case Constants.uuids.ATT_UUID_USER_DATA_SERVICE:
                        name = "User Data";
                        break;

                    case Constants.uuids.ATT_UUID_IP_SUPPORT_SERVICE:
                        name = "IP Support";
                        break;

                    case Constants.uuids.ATT_UUID_PULSE_OXIMITER_SERVICE:
                        name = "Pulse Oximeter";
                        break;

                    default:
                        name = "Unknown";
                        break;
                }
            }
            else
            {
                byte[] wsruOtaServiceUUID = { 0x6F, 0x38, 0xA1, 0x38, 0xAD, 0x82, 0x35, 0x86, 0xA0, 0x43, 0x13, 0x5C, 0x47, 0x1E, 0x5D, 0x9E };
                byte[] qorvoStrServiceUUID = { 0x01, 0x10, 0xA5, 0x4D, 0x57, 0x20, 0x4F, 0xA3, 0x50, 0x4D, 0x70, 0x85, 0x56, 0x6F, 0x91, 0x0D };
                byte[] ueiUugServiceUUID = { 0x77, 0xAE, 0x8C, 0x12, 0x71, 0x9E, 0x7B, 0xB6, 0xE6, 0x11, 0x3A, 0x21, 0x00, 0xE1, 0x41, 0x21 };

                svcString = string.Format("UUID: {0:X} - ", BitConverter.ToString(uuid.UUID128));

                if (uuid.UUID128.SequenceEqual(wsruOtaServiceUUID))
                {
                    name = "WSRU OTA";
                }
                else if (uuid.UUID128.SequenceEqual(qorvoStrServiceUUID))
                {
                    name = "Qorvo Streaming Data";
                }
                else if (uuid.UUID128.SequenceEqual(ueiUugServiceUUID))
                {
                    name = "UEI UUG";
                }
                else
                {
                    name = "Unknown";
                }
            }
            return svcString + name + " Service";
        }

        private static string GetProperties(CyGattCharacteristicProperty props)
        {
            string propString = string.Format("Properties: 0x{0:X} - ", props);

            if (props == CyGattCharacteristicProperty.NONE)
            {
                propString += "NONE";
            }
            else
            {
                if (props.HasFlag(CyGattCharacteristicProperty.BROADCAST))
                {
                    propString += "Broadcast, ";
                }
                if (props.HasFlag(CyGattCharacteristicProperty.READ))
                {
                    propString += "Read, ";
                }
                if (props.HasFlag(CyGattCharacteristicProperty.WRITE_WITHOUT_RESPONSE))
                {
                    propString += "Write w/o response, ";
                }
                if (props.HasFlag(CyGattCharacteristicProperty.WRITE))
                {
                    propString += "Write, ";
                }
                if (props.HasFlag(CyGattCharacteristicProperty.NOTIFY))
                {
                    propString += "Notify, ";
                }
                if (props.HasFlag(CyGattCharacteristicProperty.INDICATE))
                {
                    propString += "Indicate, ";
                }
                if (props.HasFlag(CyGattCharacteristicProperty.AUTHENTICATED_SIGNED_WRITES))
                {
                    propString += "Signed write, ";
                }
                if (props.HasFlag(CyGattCharacteristicProperty.EXTENDED_PROPERTIES))
                {
                    propString += "Extended, ";
                }
                // Remove the final ", " characters
                propString = propString.Remove(propString.Length - 2);
            }

            return propString;
        }

        private static string GetCharacteristicName(CyUUID uuid)
        {
            string characteristicString;
            string name;

            if (uuid.IsUUID16Valid)
            {
                characteristicString = string.Format("UUID: 0x{0:X4} - ", uuid.UUID16);
                switch ((Constants.uuids)uuid.UUID16)
                {
                    case Constants.uuids.ATT_UUID_DEVICE_NAME:
                        name = "Device Name";
                        break;

                    case Constants.uuids.ATT_UUID_APPEARANCE:
                        name = "Appearance";
                        break;

                    case Constants.uuids.ATT_UUID_PERIPH_PRIVACY_FLAG:
                        name = "Peripheral Privacy Flag";
                        break;

                    case Constants.uuids.ATT_UUID_RECONN_ADDR:
                        name = "Reconnection Address";
                        break;

                    case Constants.uuids.ATT_UUID_PREF_CONN_PARAM:
                        name = "Peripheral Preferred Connection Parameters";
                        break;

                    case Constants.uuids.ATT_UUID_SERVICE_CHANGED:
                        name = "Service Changed";
                        break;

                    case Constants.uuids.ATT_UUID_ALERT_LEVEL:
                        name = "Alert Level";
                        break;

                    case Constants.uuids.ATT_UUID_TX_POWER_LEVEL:
                        name = "TX Power Level";
                        break;

                    case Constants.uuids.ATT_UUID_DATE_TIME:
                        name = "Date Time";
                        break;

                    case Constants.uuids.ATT_UUID_DAY_OF_WEEK:
                        name = "Day of Week";
                        break;

                    case Constants.uuids.ATT_UUID_DAY_DATE_TIME:
                        name = "Day Date Time";
                        break;

                    case Constants.uuids.ATT_UUID_EXACT_TIME_100:
                        name = "Exact Time 100";
                        break;

                    case Constants.uuids.ATT_UUID_EXACT_TIME_256:
                        name = "Exact Time 256";
                        break;

                    case Constants.uuids.ATT_UUID_DST_OFFSET:
                        name = "DST Offset";
                        break;

                    case Constants.uuids.ATT_UUID_TIME_ZONE:
                        name = "Time Zone";
                        break;

                    case Constants.uuids.ATT_UUID_LOCAL_TIME_INFO:
                        name = "Local Time Information";
                        break;

                    case Constants.uuids.ATT_UUID_SECONDARY_TIME_ZONE:
                        name = "Secondary Time Zone";
                        break;

                    case Constants.uuids.ATT_UUID_TIME_WITH_DST:
                        name = "Time with DST";
                        break;

                    case Constants.uuids.ATT_UUID_TIME_ACCURACY:
                        name = "Time Accuracy";
                        break;

                    case Constants.uuids.ATT_UUID_TIME_SOURCE:
                        name = "Time Source";
                        break;

                    case Constants.uuids.ATT_UUID_REFERENCE_TIME_INFO:
                        name = "Reference Time Information";
                        break;

                    case Constants.uuids.ATT_UUID_TIME_BROADCAST:
                        name = "Time Broadcast";
                        break;

                    case Constants.uuids.ATT_UUID_TIME_UPDATE_CP:
                        name = "Time Update Control Point";
                        break;

                    case Constants.uuids.ATT_UUID_TIME_UPDATE_STATE:
                        name = "Time Update State";
                        break;

                    case Constants.uuids.ATT_UUID_GLUCOSE_MEAS:
                        name = "Glucose Measurement";
                        break;

                    case Constants.uuids.ATT_UUID_BATTERY_LEVEL:
                        name = "Battery Level";
                        break;

                    case Constants.uuids.ATT_UUID_BATTERY_POWER_STATE:
                        name = "Battery Power State";
                        break;

                    case Constants.uuids.ATT_UUID_BATTERY_LEVEL_STATE:
                        name = "Battery Level State";
                        break;

                    case Constants.uuids.ATT_UUID_TEMP_MEAS:
                        name = "Temoerature Measurement";
                        break;

                    case Constants.uuids.ATT_UUID_TEMP_TYPE:
                        name = "Temperature Type";
                        break;

                    case Constants.uuids.ATT_UUID_INTERMEDIATE_TEMP:
                        name = "Intermediate Temperature";
                        break;

                    case Constants.uuids.ATT_UUID_TEMP_C:
                        name = "Temperature Celsius";
                        break;

                    case Constants.uuids.ATT_UUID_TEMP_F:
                        name = "Temperature Fahrenheit";
                        break;

                    case Constants.uuids.ATT_UUID_MEAS_INTERVAL:
                        name = "Measurement Interval";
                        break;

                    case Constants.uuids.ATT_UUID_HID_BOOT_KEYBOARD_IN:
                        name = "HID Boot Keyboard In";
                        break;

                    case Constants.uuids.ATT_UUID_SYSTEM_ID:
                        name = "System ID";
                        break;

                    case Constants.uuids.ATT_UUID_MODEL_NUMBER:
                        name = "Model Number String";
                        break;

                    case Constants.uuids.ATT_UUID_SERIAL_NUMBER:
                        name = "Serial Number String";
                        break;

                    case Constants.uuids.ATT_UUID_FIRMWARE_REV:
                        name = "Firmware Revision String";
                        break;

                    case Constants.uuids.ATT_UUID_HARDWARE_REV:
                        name = "Hardware Revision String";
                        break;

                    case Constants.uuids.ATT_UUID_SOFTWARE_REV:
                        name = "Software Revision String";
                        break;

                    case Constants.uuids.ATT_UUID_MANUFACTURER_NAME:
                        name = "Manufacturer Name String";
                        break;

                    case Constants.uuids.ATT_UUID_11073_CERT_DATA:
                        name = "IEEE 11073-20601 Regulatory Certification Data List";
                        break;

                    case Constants.uuids.ATT_UUID_CURRENT_TIME:
                        name = "Current Time";
                        break;

                    case Constants.uuids.ATT_UUID_ELEVATION:
                        name = "Elevation";
                        break;

                    case Constants.uuids.ATT_UUID_LATITUDE:
                        name = "Latitude";
                        break;

                    case Constants.uuids.ATT_UUID_LONGITUDE:
                        name = "Longitude";
                        break;

                    case Constants.uuids.ATT_UUID_POSITION_2D:
                        name = "Position 2D";
                        break;

                    case Constants.uuids.ATT_UUID_POSITION_3D:
                        name = "Position 3D";
                        break;

                    case Constants.uuids.ATT_UUID_VENDOR_ID:
                        name = "Vendor ID";
                        break;

                    case Constants.uuids.ATT_UUID_HID_BOOT_KEYBOARD_OUT:
                        name = "HID Boot Keyboard Out";
                        break;

                    case Constants.uuids.ATT_UUID_HID_BOOT_MOUSE_IN:
                        name = "HID Boot Mouse In";
                        break;

                    case Constants.uuids.ATT_UUID_GLUCOSE_MEAS_CONTEXT:
                        name = "Glucose Measurement Context";
                        break;

                    case Constants.uuids.ATT_UUID_BP_MEAS:
                        name = "Blood Pressure Measurement";
                        break;

                    case Constants.uuids.ATT_UUID_INTERMEDIATE_BP:
                        name = "Intermediate Cuff Pressure";
                        break;

                    case Constants.uuids.ATT_UUID_HR_MEAS:
                        name = "Heart Rate Measurement";
                        break;

                    case Constants.uuids.ATT_UUID_HR_SENSOR_LOC:
                        name = "Body Sensor Location";
                        break;

                    case Constants.uuids.ATT_UUID_HR_CP:
                        name = "Heart Rate Control Point";
                        break;

                    case Constants.uuids.ATT_UUID_REMOVABLE:
                        name = "Removable";
                        break;

                    case Constants.uuids.ATT_UUID_SERVICE_REQ:
                        name = "Service Required";
                        break;

                    case Constants.uuids.ATT_UUID_SCI_TEMP_C:
                        name = "Scientific Temperature in Celsius";
                        break;

                    case Constants.uuids.ATT_UUID_STRING:
                        name = "String";
                        break;

                    case Constants.uuids.ATT_UUID_NETWORK_AVAIL:
                        name = "Network Availability";
                        break;

                    case Constants.uuids.ATT_UUID_ALERT_STATUS:
                        name = "Alert Status";
                        break;

                    case Constants.uuids.ATT_UUID_RINGER_CP:
                        name = "Ringer Control Point";
                        break;

                    case Constants.uuids.ATT_UUID_RINGER_SETTING:
                        name = "Ringer Setting";
                        break;

                    case Constants.uuids.ATT_UUID_ALERT_CAT_ID_MASK:
                        name = "Alert Category ID Bit Mask";
                        break;

                    case Constants.uuids.ATT_UUID_ALERT_CAT_ID:
                        name = "Alert Category ID";
                        break;

                    case Constants.uuids.ATT_UUID_ALERT_NOTIF_CP:
                        name = "Alert Notification Control Point";
                        break;

                    case Constants.uuids.ATT_UUID_UNREAD_ALERT_STATUS:
                        name = "Unread Alert Status";
                        break;

                    case Constants.uuids.ATT_UUID_NEW_ALERT:
                        name = "New Alert";
                        break;

                    case Constants.uuids.ATT_UUID_SUP_NEW_ALERT_CAT:
                        name = "Supported New Alert Category";
                        break;

                    case Constants.uuids.ATT_UUID_SUP_UNREAD_ALERT_CAT:
                        name = "Supported Unread Alert Category";
                        break;

                    case Constants.uuids.ATT_UUID_BP_FEATURE:
                        name = "Blood Pressure Feature";
                        break;

                    case Constants.uuids.ATT_UUID_HID_INFORMATION:
                        name = "HID Information";
                        break;

                    case Constants.uuids.ATT_UUID_HID_REPORT_MAP:
                        name = "HID Report Map";
                        break;

                    case Constants.uuids.ATT_UUID_HID_CONTROL_POINT:
                        name = "HID Control Point";
                        break;

                    case Constants.uuids.ATT_UUID_HID_REPORT:
                        name = "HID Report";
                        break;

                    case Constants.uuids.ATT_UUID_HID_PROTOCOL_MODE:
                        name = "HID Protocol Mode";
                        break;

                    case Constants.uuids.ATT_UUID_SCAN_INT_WIND:
                        name = "Scan Interval Window";
                        break;

                    case Constants.uuids.ATT_UUID_PNP_ID:
                        name = "Pnp ID";
                        break;

                    case Constants.uuids.ATT_UUID_GLUCOSE_FEATURE:
                        name = "Glucose Feature";
                        break;

                    case Constants.uuids.ATT_UUID_RACP:
                        name = "Record Access Control Point";
                        break;

                    case Constants.uuids.ATT_UUID_RUNNING_SPEED_MEASUREMENT:
                        name = "Running Speed Measurement";
                        break;

                    case Constants.uuids.ATT_UUID_RUNNING_SPEED_FEATURE:
                        name = "Running Speed Feature";
                        break;

                    case Constants.uuids.ATT_UUID_CYCLING_SPEED_MEASUREMENT:
                        name = "Cycling Speed Feature";
                        break;

                    case Constants.uuids.ATT_UUID_CYCLING_SPEED_FEATURE:
                        name = "Cycling Speed Measurement";
                        break;

                    case Constants.uuids.ATT_UUID_SENSOR_LOCATION:
                        name = "Sensor Location";
                        break;

                    case Constants.uuids.ATT_UUID_PULSE_OX_SPOT_CHECK:
                        name = "Pulse Oximeter Spot Check";
                        break;

                    case Constants.uuids.ATT_UUID_PULSE_OX_CONTINUOUS:
                        name = "Pulse Oximeter Continuous";
                        break;

                    case Constants.uuids.ATT_UUID_PULSE_OX_FEATURES:
                        name = "Pulse Oximeter Features";
                        break;

                    case Constants.uuids.ATT_UUID_CYCLING_POWER_MEASUREMENT:
                        name = "Cycling Power Measurement";
                        break;

                    case Constants.uuids.ATT_UUID_CYCLING_POWER_FEATURE:
                        name = "Cycling Power Feature";
                        break;

                    case Constants.uuids.ATT_UUID_DB_CHANGE_INCREMENT:
                        name = "Database Change Increment";
                        break;

                    case Constants.uuids.ATT_UUID_USER_INDEX:
                        name = "User Index";
                        break;

                    case Constants.uuids.ATT_UUID_USER_CONTROL_POINT:
                        name = "User Control Point";
                        break;

                    case Constants.uuids.ATT_UUID_CAR:
                        name = "Central Address Resolution";
                        break;

                    case Constants.uuids.ATT_UUID_RPAO:
                        name = "Resolvable Private Address Only";
                        break;

                    default:
                        name = "Unknown";
                        break;
                }
            }
            else
            {
                byte[] wsruOtaControlPointCharacteristicUUID =
                    { 0x6B, 0x66, 0x6C, 0x08, 0x0A, 0x57, 0x8E, 0x83, 0x99, 0x4E, 0xA7, 0xF7, 0xBF, 0x50, 0xDD, 0xE3 };
                byte[] wsruOtaDataCharacteristicUUID =
                    { 0x36, 0xFE, 0x2E, 0xE7, 0x09, 0x24, 0x4F, 0xB7, 0x91, 0x40, 0x61, 0xD9, 0x7A, 0x6C, 0xE8, 0x92 };
                byte[] qorvoStrControlPointCharacteristicUUID =
                    { 0x02, 0x00, 0xA5, 0x4D, 0x57, 0x20, 0x4F, 0xA3, 0x50, 0x4D, 0x70, 0x85, 0x56, 0x6F, 0x91, 0x0D };
                byte[] qorvoStrDataCharacteristicUUID =
                    { 0x01, 0x00, 0xA5, 0x4D, 0x57, 0x20, 0x4F, 0xA3, 0x50, 0x4D, 0x70, 0x85, 0x56, 0x6F, 0x91, 0x0D };
                byte[] ueiUugUapiInputDataUUID =
                    { 0x77, 0xAE, 0x8C, 0x12, 0x71, 0x9E, 0x7B, 0xB6, 0xE6, 0x11, 0x3A, 0x21, 0x01, 0xE1, 0x41, 0x21 };
                byte[] ueiUugUapiOutputDataUUID =
                    { 0x77, 0xAE, 0x8C, 0x12, 0x71, 0x9E, 0x7B, 0xB6, 0xE6, 0x11, 0x3A, 0x21, 0x02, 0xE1, 0x41, 0x21 };
                byte[] ueiUugVoiceInputDataUUID =
                    { 0x77, 0xAE, 0x8C, 0x12, 0x71, 0x9E, 0x7B, 0xB6, 0xE6, 0x11, 0x3A, 0x21, 0x03, 0xE1, 0x41, 0x21 };

                characteristicString = string.Format("UUID: {0:X} - ", BitConverter.ToString(uuid.UUID128));

                if (uuid.UUID128.SequenceEqual(wsruOtaControlPointCharacteristicUUID))
                {
                    name = "WSRU OTA Service Control Point";
                }
                else if (uuid.UUID128.SequenceEqual(wsruOtaDataCharacteristicUUID))
                {
                    name = "WSRU OTA Service Data";
                }
                else if (uuid.UUID128.SequenceEqual(qorvoStrControlPointCharacteristicUUID))
                {
                    name = "Qorvo Streaming Data Svc Control Point";
                }
                else if (uuid.UUID128.SequenceEqual(qorvoStrDataCharacteristicUUID))
                {
                    name = "Qorvo Streaming Data Svc Data";
                }
                else if (uuid.UUID128.SequenceEqual(ueiUugUapiInputDataUUID))
                {
                    name = "UEI UUG Service UAPI Input Data";
                }
                else if (uuid.UUID128.SequenceEqual(ueiUugUapiOutputDataUUID))
                {
                    name = "UEI UUG Service UAPI Output Data";
                }
                else if (uuid.UUID128.SequenceEqual(ueiUugVoiceInputDataUUID))
                {
                    name = "UEI UUG Service Voice Input Data";
                }
                else
                {
                    name = "Unknown";
                }
            }
            return characteristicString + name;
        }

        private static string GetDescriptorName(CyUUID uuid)
        {
            string descriptorString;
            string name;

            if (uuid.IsUUID16Valid)
            {
                descriptorString = string.Format("UUID: 0x{0:X4} - ", uuid.UUID16);
                switch ((Constants.uuids)uuid.UUID16)
                {
                    case Constants.uuids.ATT_UUID_CHARACTERISTIC_EXT:
                        name = "Characteristic Extended Properties";
                        break;

                    case Constants.uuids.ATT_UUID_CHAR_USER_DESC:
                        name = "Characteristic User Description";
                        break;

                    case Constants.uuids.ATT_UUID_CLIENT_CHAR_CONFIG:
                        name = "Client Characteristic Configuration";
                        break;

                    case Constants.uuids.ATT_UUID_SERVER_CHAR_CONFIG:
                        name = "Server Characteristic Configuration";
                        break;

                    case Constants.uuids.ATT_UUID_CHAR_PRES_FORMAT:
                        name = "Characteristic Presentation Format";
                        break;

                    case Constants.uuids.ATT_UUID_AGGREGATE_FORMAT:
                        name = "Characteristic Aggregate Format";
                        break;

                    case Constants.uuids.ATT_UUID_VALID_RANGE:
                        name = "Valid Range";
                        break;

                    case Constants.uuids.ATT_UUID_HID_EXT_REPORT_MAPPING:
                        name = "HID External Report ID Mapping";
                        break;

                    case Constants.uuids.ATT_UUID_HID_REPORT_ID_MAPPING:
                        name = "HID Report ID Mapping";
                        break;

                    default:
                        name = "Unknown";
                        break;
                }
            }
            else
            {
                descriptorString = string.Format("UUID: {0:X} - ", BitConverter.ToString(uuid.UUID128));
                name = "Unknown";
            }
            return descriptorString + name;
        }

        public static void InitializeHID()
        {
            Globals.hidState = Constants.hidStates.HID_STATE_INCOMPLETE;

            foreach (var svc in Program.mf.gattData.Services)
            {
                // Have we found the HID Service?
                if (svc.UUID.IsUUID16Valid && ((Constants.uuids)svc.UUID.UUID16 == Constants.uuids.ATT_UUID_HID_SERVICE))
                {
                    Globals.uapiHidData.inputIdx = -1;
                    Globals.uapiHidData.outputIdx = -1;
                    Globals.uapiHidData.inputs.Clear();
                    Globals.uapiHidData.outputs.Clear();

                    foreach (var chr in svc.Characteristics)
                    {
                        // Have we found a HID Report?
                        if (chr.UUID.IsUUID16Valid && ((Constants.uuids)chr.UUID.UUID16 == Constants.uuids.ATT_UUID_HID_REPORT))
                        {
                            Globals.hidReportChar rptChar = new Globals.hidReportChar { };

                            // Does it appear to be a valid Output Report?
                            if ( (chr.Descriptors.Count == 1) &&
                                 chr.Properties.HasFlag(CyGattCharacteristicProperty.WRITE_WITHOUT_RESPONSE) &&
                                 chr.Descriptors[0].UUID.IsUUID16Valid && ((Constants.uuids)chr.Descriptors[0].UUID.UUID16 == Constants.uuids.ATT_UUID_HID_REPORT_ID_MAPPING))
                            {
                                rptChar.mappingHdl = chr.Descriptors[0].Handle;
                                rptChar.dataHdl = chr.Handle;
                                Globals.uapiHidData.outputs.Add(rptChar);
                            }
                            // Does it appear to be a valid Input Report?
                            else if ( (chr.Descriptors.Count == 2) &&
                                      chr.Properties.HasFlag(CyGattCharacteristicProperty.READ | CyGattCharacteristicProperty.NOTIFY) &&
                                      chr.Descriptors[0].UUID.IsUUID16Valid && chr.Descriptors[1].UUID.IsUUID16Valid)
                            {
                                if ( ((Constants.uuids)chr.Descriptors[0].UUID.UUID16 == Constants.uuids.ATT_UUID_HID_REPORT_ID_MAPPING) &&
                                     ((Constants.uuids)chr.Descriptors[1].UUID.UUID16 == Constants.uuids.ATT_UUID_CLIENT_CHAR_CONFIG) )
                                {
                                    rptChar.mappingHdl = chr.Descriptors[0].Handle;
                                    rptChar.cccHdl = chr.Descriptors[1].Handle;
                                }
                                else if ( ((Constants.uuids)chr.Descriptors[1].UUID.UUID16 == Constants.uuids.ATT_UUID_HID_REPORT_ID_MAPPING) &&
                                          ((Constants.uuids)chr.Descriptors[0].UUID.UUID16 == Constants.uuids.ATT_UUID_CLIENT_CHAR_CONFIG) )
                                {
                                    rptChar.mappingHdl = chr.Descriptors[1].Handle;
                                    rptChar.cccHdl = chr.Descriptors[0].Handle;
                                }
                                else
                                {
                                    return;
                                }
                                rptChar.dataHdl = chr.Handle;
                                Globals.uapiHidData.inputs.Add(rptChar);
                            }
                            else
                            {
                                return;
                            }
                        }
                    }
                }
            }

            if ( (Globals.uapiHidData.inputs.Count != 0) && (Globals.uapiHidData.outputs.Count != 0) )
            {
                Globals.hidState = Constants.hidStates.HID_STATE_CONFIRMING_INPUT;
                Uapi.uapiMessages.AppendText("Confirming Inputs\n", Color.Blue);

                Globals.uapiHidData.mappingHdlIdx = 0;
                CyGattReadInfo info = new CyGattReadInfo(Globals.uapiHidData.inputs[0].mappingHdl);
                CyApiErr err = Program.mf.gattClient.ReadCharacteristic(info);
                if (err.IsNotOK)
                {
#if (debugTestGatt)
                    // The following is debug-only nonsense
                    Utilities.ForceException(25);
#endif
                }
            }
        }
    }
}
