﻿#undef debugTestUapi
#define debugTestUapi

using CySmart.Common.Base.Compatibility.Net20;
using CySmart.DongleCommunicator.API;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO.Ports;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;


namespace Host
{
    public static class Uapi
    {
        // The overall UAPI dialog box itself
        private static Form uapiBox;
        // A container for the button block, parameter listview, and two message boxes
        private static TableLayoutPanel panel;
        // This shows the cumulative UAPI messages between the Host and the remote as ASCII hex bytes.
        // Note that this is public so that CySmart callbacks can update the box with response messages.
        public static RichTextBox uapiMessages;
        // This shows a manually entered UAPI message (for custom messages)
        private static TextBox newMessage;
        // This contains the various UAPI selection and command buttons
        private static TableLayoutPanel uapiButtons;
        // Two radio buttons - for selecting either AppID 0x00 or AppID 0x13
        private static RadioButton appID00;
        private static RadioButton appID13;
        // Buttons specific to App ID 0x00
        private static Button queryFeature;
        private static Button queryApp;
        private static Button echo;
        // A collection of buttons that facilitates search and classification
        private static List<Button> appID00Buttons;
        // Buttons specific to App ID 0x13
        private static Button startSession;
        private static Button stopSession;
        private static Button startDiscovery;
        private static Button stopDiscovery;
        private static Button setCfg;
        private static Button writeRecord;
        private static Button deleteRecord;
        private static Button readRecord;
        private static Button listRecords;
        private static Button residentSetup;
        // A collection of buttons that facilitates search and classification
        private static List<Button> appID13Buttons;
        // This listview presents the configuration parameters for the currently selected AppID
        private static ListView cfgParameters;
        // Version Info - UAPI version information (Version X.Y.Z is encoded as 0xXX, 0xYY, 0xZZZZ)
        private static ListViewItem appID00_param01;
        // Product Info - product information comprising 4-byte platform version and 10-character (byte) product name
        private static ListViewItem appID00_param02;
        // Polling Period - period used by app to poll for messages
        private static ListViewItem appID00_param03;
        // HW Info - HW platform information; 11 bytes (4 single-byte values plus 7 filler bytes) word-aligned up to 12 bytes
        private static ListViewItem appID00_param04;
        // Auto Poll State = automatic polling enable/disable
        private static ListViewItem appID00_param05;
        // Bootloader Version Info - bootloader version information (Version X.Y.Z is encoded as 0xXX, 0xYY, 0xZZZZ)
        private static ListViewItem appID00_param06;
        // Max Payload - maximum supported payload size in bytes of incoming and outgoing UAPI messages; defaults to 1020
        private static ListViewItem appID00_param07;
        // Secured - true if network used to transfer UAPI messages is secured; if not present, assume false
        private static ListViewItem appID00_param08;
        // MTU Size - MTU size in bytes; if not present, assume 20 bytes
        private static ListViewItem appID00_param09;
        // A collection of listview items that facilitates search and classification
        private static List<ListViewItem> appID00params;
        // Enable - enable/disable QuickSet
        private static ListViewItem appID13_param00;
        // Flags - b7: set if RF is supported; b6: set if IR is supported; b5-b0: reserved, must be 0
        private static ListViewItem appID13_param01;
        // SupportedEncryptionMethods - encryption methods supported by the RCU; bit n corresponds to encryption method n+1
        private static ListViewItem appID13_param02;
        // EncryptionMethod - encryption method to be used on all payloads; 0 = no encryption, 1 = AES-128
        private static ListViewItem appID13_param03;
        // NumberOfRFSlots - total number of RF configuration slots; if RF is supported, this value must be at least 1
        private static ListViewItem appID13_param04;
        // SessionTimeout - session timeout in milliseconds; an open session will automatically close 'SessionTimeout' ms
        //  after receiving the last QS message
        private static ListViewItem appID13_param05;
        // MaxRecordSize - maximum record payload size supported
        private static ListViewItem appID13_param06;
        // MinRecordID - minimum record ID available
        private static ListViewItem appID13_param07;
        // MaxRecordID - maximum record ID available
        private static ListViewItem appID13_param08;
        // ActiveKeyMap - record ID of active key map; 0x0000 indicates that no key map is active
        private static ListViewItem appID13_param09;
        // RF control module version information; version X.Y.Z is encoded as X (1 byte), Y (1 byte), Z (2 bytes)
        private static ListViewItem appID13_param0A;
        // A collection of listview items that facilitates search and classification
        private static List<ListViewItem> appID13params;

        // The overall parameter input dialog box itself
        private static Form paramBox;
        // A container for the action button block, location button block, value textbox, and standard window buttons
        private static TableLayoutPanel paramPanel;
        // This shows a manually entered parameter value to be written (set)
        private static TableLayoutPanel setValue;
        private static Label valueLabel;
        private static TextBox setValueBox;
        // Three radio buttons - for selecting the desired Action
        private static TableLayoutPanel actions;
        private static Label actionLabel;
        private static RadioButton get;
        private static RadioButton set;
        private static RadioButton copy;
        // Three radio buttons - for selecting the Location
        private static TableLayoutPanel locations;
        private static Label locationLabel;
        private static RadioButton ram;
        private static RadioButton nvm;
        private static RadioButton def;
        private static TableLayoutPanel windowButtons;
        // The usual 'Ok' and 'Cancel' window buttons
        private static Button ok;
        private static Button cancel;

        private static byte opCode;
        private static byte action;
        private static byte uapiSeqNum;

        public static byte txFragCnt;
        public static byte[] msgPayload;
        public static ushort msgPayloadLen;
        private static ushort msgIndex;
        private static byte msgAppID;
        private static byte msgOpCode;

        private class uapiButtonObject : Object
        {
            public byte appID { get; set; }
            public int row { get; set; }
            public int col { get; set; }

            public uapiButtonObject(byte id, int r, int c)
            {
                appID = id;
                row = r;
                col = c;
            }
        }

        private class uapiParameterObject : Object
        {
            public byte aID { get; set; }
            public byte pID { get; set; }
            public byte dType { get; set; }
            public int len { get; set; }
            public byte[] value { get; set; }
            public Constants.configParameterAccess rw { get; set; }
            public uapiParameterObject(byte appID, byte paramID, byte dataType, int length, Constants.configParameterAccess access)
            {
                aID = appID;
                pID = paramID;
                dType = dataType;
                len = length;
                rw = access;
            }
        }

        static Uapi()
        {
            // First, define the overall UAPI dialog box itself
            uapiBox = new Form();
            uapiBox.Name = "uapiBox";
            uapiBox.Width = 1000;
            uapiBox.Height = 800;
            uapiBox.Text = "UAPI Messaging";
            uapiBox.Icon = Program.mf.Icon;
            uapiBox.FormClosed += (sender, e) =>
            {
                Utilities.SetState(Constants.hostStates.HOST_STATE_DEVICE_CONNECTED);
            };

            // Define a container for the button block, parameter listview, and two message boxes
            panel = new TableLayoutPanel()
            {
                ColumnCount = 2,
                Dock = System.Windows.Forms.DockStyle.Fill,
                Name = "panel",
                RowCount = 3,
                TabIndex = 0
            };
            panel.ColumnStyles.Add(new System.Windows.Forms.ColumnStyle(System.Windows.Forms.SizeType.Percent, 50F));
            panel.ColumnStyles.Add(new System.Windows.Forms.ColumnStyle(System.Windows.Forms.SizeType.Percent, 50F));
            panel.RowStyles.Add(new System.Windows.Forms.RowStyle(System.Windows.Forms.SizeType.Percent, 47.5F));
            panel.RowStyles.Add(new System.Windows.Forms.RowStyle(System.Windows.Forms.SizeType.Percent, 47.5F));
            panel.RowStyles.Add(new System.Windows.Forms.RowStyle(System.Windows.Forms.SizeType.Percent, 5F));

            // This shows the cumulative UAPI messages between the Host and the remote as ASCII hex bytes
            uapiMessages = new RichTextBox()
            {
                Dock = System.Windows.Forms.DockStyle.Fill,
                ForeColor = Color.Green,
                //Width = 600,
                //Height = 720
            };

            // This shows a manually entered UAPI message (for custom messages). Fragmentation is not supported!
            newMessage = new TextBox()
            {
                Dock = System.Windows.Forms.DockStyle.Fill
            };
            newMessage.Enter += (sender, e) =>
            {
                newMessage.Clear();
            };
            newMessage.KeyUp += (sender, e) =>
            {
                if (e.KeyCode == Keys.Enter)
                {
                    uapiMessages.AppendText(newMessage.Text + "\n", Color.Green);
                    byte[] messageData = Utilities.HexStringToByteArray(newMessage.Text, 0);
                    if ( (messageData != null) && (messageData.Length >= 2) )
                    {
                        msgPayloadLen = (ushort)(messageData.Length - 2);
                        Array.Copy(messageData, 2, msgPayload, 0, msgPayloadLen);
                        SendMessage(messageData[0], messageData[1]);
                    }

                    newMessage.Clear();
                }
            };

            // This contains the various UAPI selection and command buttons
            uapiButtons = new TableLayoutPanel()
            {
                ColumnCount = 2,
                Dock = System.Windows.Forms.DockStyle.Fill,
                Name = "uapiButtons",
                RowCount = 6
            };

            uapiButtons.ColumnStyles.Add(new System.Windows.Forms.ColumnStyle(System.Windows.Forms.SizeType.Percent, 50F));
            uapiButtons.ColumnStyles.Add(new System.Windows.Forms.ColumnStyle(System.Windows.Forms.SizeType.Percent, 50F));
            for (int i = 0; i < 6; i++)
            {
                uapiButtons.RowStyles.Add(new System.Windows.Forms.RowStyle(System.Windows.Forms.SizeType.Percent, 16.7F));
            }

            // Two radio buttons - for selecting either AppID 0x00 or AppID 0x13
            appID00 = new RadioButton()
            {
                Checked = false,
                Text = "App ID 0x00",
                Tag = new uapiButtonObject(0xFF, 0, 0)
            };

            appID13 = new RadioButton()
            {
                Checked = true,
                Text = "App ID 0x13",
                Tag = new uapiButtonObject(0xFF, 0, 1)
            };

            // Buttons specific to App ID 0x00
            queryFeature = new Button()
            {
                Text = "Query Feature (0x44)",
                Width = 150,
                Height = 23,
                Tag = new uapiButtonObject(0x00, 1, 0)
            };
            queryFeature.Click += (sender, e) =>
            {
                uapiMessages.AppendText(queryFeature.Text, Color.Blue);
            };

            queryApp = new Button()
            {
                Text = "Query App (0x44)",
                Width = 150,
                Height = 23,
                Tag = new uapiButtonObject(0x00, 1, 1)
            };
            queryApp.Click += (sender, e) =>
            {
                uapiMessages.AppendText(queryApp.Text, Color.Blue);
            };

            echo = new Button()
            {
                Text = "Echo (0x42)",
                Width = 150,
                Height = 23,
                Tag = new uapiButtonObject(0x00, 2, 0)
            };
            echo.Click += (sender, e) =>
            {
                uapiMessages.AppendText(echo.Text, Color.Blue);
            };

            // A collection of buttons that facilitates search and classification
            appID00Buttons = new List<Button>
            {
                queryFeature, queryApp, echo
            };

            // Buttons specific to App ID 0x13
            startSession = new Button()
            {
                Text = "Start Session (0x41)",
                Width = 150,
                Height = 23,
                Tag = new uapiButtonObject(0x13, 1, 0)
            };
            startSession.Click += (sender, e) =>
            {
                if (txFragCnt == 0)
                {
                    msgPayloadLen = 0;
                    SendMessage(0x13, 0x41);
                }
            };

            stopSession = new Button()
            {
                Text = "Stop Session (0x42)",
                Width = 150,
                Height = 23,
                Tag = new uapiButtonObject(0x13, 1, 1)
            };
            stopSession.Click += (sender, e) =>
            {
                if (txFragCnt == 0)
                {
                    msgPayloadLen = 0;
                    SendMessage(0x13, 0x42);
                }
            };

            startDiscovery = new Button()
            {
                Text = "Start Discovery (0x43)",
                Width = 150,
                Height = 23,
                Tag = new uapiButtonObject(0x13, 2, 0)
            };
            startDiscovery.Click += (sender, e) =>
            {
                Rfdb.RfdbDialogBox(false);
            };

            stopDiscovery = new Button()
            {
                Text = "Stop Discovery (0x44)",
                Width = 150,
                Height = 23,
                Tag = new uapiButtonObject(0x13, 2, 1)
            };
            stopDiscovery.Click += (sender, e) =>
            {
                if (txFragCnt == 0)
                {
                    msgPayloadLen = 0;
                    SendMessage(0x13, 0x44);
                }
            };

            setCfg = new Button()
            {
                Text = "Set Config (0x47)",
                Width = 150,
                Height = 23,
                Tag = new uapiButtonObject(0x13, 3, 0)
            };
            setCfg.Click += (sender, e) =>
            {
                DialogResult result = Rfdb.RfdbGetSlot();

                if (result == DialogResult.OK)
                {
                    Rfdb.RfdbDialogBox(true);
                }
            };

            writeRecord = new Button()
            {
                Text = "Write Record (0x4A)",
                Width = 150,
                Height = 23,
                Tag = new uapiButtonObject(0x13, 3, 1)
            };
            writeRecord.Click += (sender, e) =>
            {
                uapiMessages.AppendText(writeRecord.Text + "\n", Color.Brown);
            };

            deleteRecord = new Button()
            {
                Text = "Delete Record (0x4B)",
                Width = 150,
                Height = 23,
                Tag = new uapiButtonObject(0x13, 4, 0)
            };
            deleteRecord.Click += (sender, e) =>
            {
                uapiMessages.AppendText(deleteRecord.Text + "\n", Color.Brown);
            };

            readRecord = new Button()
            {
                Text = "Read Record (0x4C)",
                Width = 150,
                Height = 23,
                Tag = new uapiButtonObject(0x13, 4, 1)
            };
            readRecord.Click += (sender, e) =>
            {
                uapiMessages.AppendText(readRecord.Text + "\n", Color.Brown);
            };

            listRecords = new Button()
            {
                Text = "List All Records (0x4E)",
                Width = 150,
                Height = 23,
                Tag = new uapiButtonObject(0x13, 5, 0)
            };
            listRecords.Click += (sender, e) =>
            {
                uapiMessages.AppendText(listRecords.Text + "\n", Color.Brown);
            };

            residentSetup = new Button()
            {
                Text = "Resident Setup (0x50)",
                Width = 150,
                Height = 23,
                Tag = new uapiButtonObject(0x13, 5, 1)
            };
            residentSetup.Click += (sender, e) =>
            {
                uapiMessages.AppendText(residentSetup.Text + "\n", Color.Brown);
            };

            // A collection of buttons that facilitates search and classification
            appID13Buttons = new List<Button>
            {
                startSession, stopSession, startDiscovery, stopDiscovery, setCfg, writeRecord, deleteRecord, readRecord,
                listRecords, residentSetup
            };

            // This listview presents the configuration parameters for the currently selected AppID
            cfgParameters = new ListView()
            {
                Dock = System.Windows.Forms.DockStyle.Fill,
                GridLines = true,
                MultiSelect = false,
                Name = "cfgParameters",
                UseCompatibleStateImageBehavior = false,
                View = System.Windows.Forms.View.Details
            };
            ColumnHeader paramID = new ColumnHeader() { Text = "Parameter ID", Width = -2 };
            ColumnHeader paramName = new ColumnHeader() { Text = "Name", Width = -1 };
            ColumnHeader paramType = new ColumnHeader() { Text = "Type", Width = -1 };
            ColumnHeader paramAccess = new ColumnHeader() { Text = "R/W", Width = -2 };
            cfgParameters.Columns.AddRange(new System.Windows.Forms.ColumnHeader[] { paramID, paramName, paramType, paramAccess });

            // Version Info - UAPI version information (Version X.Y.Z is encoded as 0xXX, 0xYY, 0xZZZZ)
            appID00_param01 = new ListViewItem("0x01");
            appID00_param01.SubItems.Add("Version Info");
            appID00_param01.SubItems.Add("Struct");
            appID00_param01.SubItems.Add("R");
            appID00_param01.Tag = new uapiParameterObject(0x00, 0x01, 0x20, 4, Constants.configParameterAccess.READ);

            // Product Info - product information comprising 4-byte platform version and 10-character (byte) product name
            appID00_param02 = new ListViewItem("0x02");
            appID00_param02.SubItems.Add("Product Info");
            appID00_param02.SubItems.Add("Struct");
            appID00_param02.SubItems.Add("R");
            appID00_param02.Tag = new uapiParameterObject(0x00, 0x02, 0x20, 14, Constants.configParameterAccess.READ);

            // Polling Period - period used by app to poll for messages
            appID00_param03 = new ListViewItem("0x03");
            appID00_param03.SubItems.Add("Polling Period");
            appID00_param03.SubItems.Add("Integer");
            appID00_param03.SubItems.Add("R/W");
            appID00_param03.Tag = new uapiParameterObject(0x00, 0x03, 0x30, 4, Constants.configParameterAccess.READ | Constants.configParameterAccess.WRITE);

            // HW Info - HW platform information; 11 bytes (4 single-byte values plus 7 filler bytes) word-aligned up to 12 bytes
            appID00_param04 = new ListViewItem("0x04");
            appID00_param04.SubItems.Add("HW Info");
            appID00_param04.SubItems.Add("Struct");
            appID00_param04.SubItems.Add("R");
            appID00_param04.Tag = new uapiParameterObject(0x00, 0x04, 0x20, 12, Constants.configParameterAccess.READ);

            // Auto Poll State = automatic polling enable/disable
            appID00_param05 = new ListViewItem("0x05");
            appID00_param05.SubItems.Add("Auto Poll Size");
            appID00_param05.SubItems.Add("Boolean");
            appID00_param05.SubItems.Add("R/W");
            appID00_param05.Tag = new uapiParameterObject(0x00, 0x05, 0x00, 1, Constants.configParameterAccess.READ | Constants.configParameterAccess.WRITE);

            // Bootloader Version Info - bootloader version information (Version X.Y.Z is encoded as 0xXX, 0xYY, 0xZZZZ)
            appID00_param06 = new ListViewItem("0x06");
            appID00_param06.SubItems.Add("Boot Loader Version Info");
            appID00_param06.SubItems.Add("Struct");
            appID00_param06.SubItems.Add("R");
            appID00_param06.Tag = new uapiParameterObject(0x00, 0x06, 0x20, 4, Constants.configParameterAccess.READ);

            // Max Payload - maximum supported payload size in bytes of incoming and outgoing UAPI messages; defaults to 1020
            appID00_param07 = new ListViewItem("0x07");
            appID00_param07.SubItems.Add("Max Payload");
            appID00_param07.SubItems.Add("Integer");
            appID00_param07.SubItems.Add("R");
            appID00_param07.Tag = new uapiParameterObject(0x00, 0x07, 0x30, 2, Constants.configParameterAccess.READ);

            // Secured - true if network used to transfer UAPI messages is secured; if not present, assume false
            appID00_param08 = new ListViewItem("0x08");
            appID00_param08.SubItems.Add("Secured");
            appID00_param08.SubItems.Add("Boolean");
            appID00_param08.SubItems.Add("R");
            appID00_param08.Tag = new uapiParameterObject(0x00, 0x08, 0x00, 1, Constants.configParameterAccess.READ);

            // MTU Size - MTU size in bytes; if not present, assume 20 bytes
            appID00_param09 = new ListViewItem("0x09");
            appID00_param09.SubItems.Add("MTU Size");
            appID00_param09.SubItems.Add("Integer");
            appID00_param09.SubItems.Add("R");
            appID00_param09.Tag = new uapiParameterObject(0x00, 0x09, 0x30, 1, Constants.configParameterAccess.READ);

            // A collection of listview items that facilitates search and classification
            appID00params = new List<ListViewItem>
            {
                appID00_param01, appID00_param02, appID00_param03, appID00_param04, appID00_param05, appID00_param06,
                appID00_param07, appID00_param08, appID00_param09
            };

            // Enable - enable/disable QuickSet
            appID13_param00 = new ListViewItem("0x00");
            appID13_param00.SubItems.Add("Enable");
            appID13_param00.SubItems.Add("Boolean");
            appID13_param00.SubItems.Add("R");
            appID13_param00.Tag = new uapiParameterObject(0x13, 0x00, 0x00, 1, Constants.configParameterAccess.READ);

            // Flags - b7: set if RF is supported; b6: set if IR is supported; b5-b0: reserved, must be 0
            appID13_param01 = new ListViewItem("0x01");
            appID13_param01.SubItems.Add("Flags");
            appID13_param01.SubItems.Add("Bit Field");
            appID13_param01.SubItems.Add("R");
            appID13_param01.Tag = new uapiParameterObject(0x13, 0x01, 0x30, 1, Constants.configParameterAccess.READ);

            // SupportedEncryptionMethods - encryption methods supported by the RCU; bit n corresponds to encryption method n+1
            appID13_param02 = new ListViewItem("0x02");
            appID13_param02.SubItems.Add("SupportedEncryptionMethods");
            appID13_param02.SubItems.Add("Bit Field");
            appID13_param02.SubItems.Add("R");
            appID13_param02.Tag = new uapiParameterObject(0x13, 0x02, 0x30, 1, Constants.configParameterAccess.READ);

            // EncryptionMethod - encryption method to be used on all payloads; 0 = no encryption, 1 = AES-128
            appID13_param03 = new ListViewItem("0x03");
            appID13_param03.SubItems.Add("EncryptionMethod");
            appID13_param03.SubItems.Add("Integer");
            appID13_param03.SubItems.Add("R/W");
            appID13_param03.Tag = new uapiParameterObject(0x13, 0x03, 0x30, 1, Constants.configParameterAccess.READ | Constants.configParameterAccess.WRITE);

            // NumberOfRFSlots - total number of RF configuration slots; if RF is supported, this value must be at least 1
            appID13_param04 = new ListViewItem("0x04");
            appID13_param04.SubItems.Add("NumberOfRFSlots");
            appID13_param04.SubItems.Add("Integer");
            appID13_param04.SubItems.Add("R");
            appID13_param04.Tag = new uapiParameterObject(0x13, 0x04, 0x30, 1, Constants.configParameterAccess.READ);

            // SessionTimeout - session timeout in milliseconds; an open session will automatically close 'SessionTimeout' ms
            //  after receiving the last QS message
            appID13_param05 = new ListViewItem("0x05");
            appID13_param05.SubItems.Add("Session Timeout");
            appID13_param05.SubItems.Add("Integer");
            appID13_param05.SubItems.Add("R/W");
            appID13_param05.Tag = new uapiParameterObject(0x13, 0x05, 0x30, 4, Constants.configParameterAccess.READ | Constants.configParameterAccess.WRITE);

            // MaxRecordSize - maximum record payload size supported
            appID13_param06 = new ListViewItem("0x06");
            appID13_param06.SubItems.Add("MaxRecordSize");
            appID13_param06.SubItems.Add("Integer");
            appID13_param06.SubItems.Add("R");
            appID13_param06.Tag = new uapiParameterObject(0x13, 0x06, 0x30, 2, Constants.configParameterAccess.READ);

            // MinRecordID - minimum record ID available
            appID13_param07 = new ListViewItem("0x07");
            appID13_param07.SubItems.Add("MinRecordID");
            appID13_param07.SubItems.Add("Integer");
            appID13_param07.SubItems.Add("R");
            appID13_param07.Tag = new uapiParameterObject(0x13, 0x07, 0x30, 2, Constants.configParameterAccess.READ);

            // MaxRecordID - maximum record ID available
            appID13_param08 = new ListViewItem("0x08");
            appID13_param08.SubItems.Add("MaxRecordID");
            appID13_param08.SubItems.Add("Integer");
            appID13_param08.SubItems.Add("R");
            appID13_param08.Tag = new uapiParameterObject(0x13, 0x08, 0x30, 2, Constants.configParameterAccess.READ);

            // ActiveKeyMap - record ID of active key map; 0x0000 indicates that no key map is active
            appID13_param09 = new ListViewItem("0x09");
            appID13_param09.SubItems.Add("ActiveKeyMap");
            appID13_param09.SubItems.Add("Integer");
            appID13_param09.SubItems.Add("R/W");
            appID13_param09.Tag = new uapiParameterObject(0x13, 0x09, 0x30, 2, Constants.configParameterAccess.READ | Constants.configParameterAccess.WRITE);

            // RF control module version information; version X.Y.Z is encoded as X (1 byte), Y (1 byte), Z (2 bytes)
            appID13_param0A = new ListViewItem("0x0A");
            appID13_param0A.SubItems.Add("RFVersion");
            appID13_param0A.SubItems.Add("Struct");
            appID13_param0A.SubItems.Add("R");
            appID13_param0A.Tag = new uapiParameterObject(0x13, 0x0A, 0x20, 4, Constants.configParameterAccess.READ);

            // A collection of listview items that facilitates search and classification
            appID13params = new List<ListViewItem>
            {
                appID13_param00, appID13_param01, appID13_param02, appID13_param03, appID13_param04, appID13_param05,
                appID13_param06, appID13_param07, appID13_param08, appID13_param09, appID13_param0A
            };

            cfgParameters.DoubleClick += (sender, e) =>
            {
                if (Globals.hidState == Constants.hidStates.HID_STATE_COMPLETE)
                {
                    ParameterDialogBox(cfgParameters.SelectedItems[0]);
                }
            };

            appID00.Click += (sender, e) =>
            {
                foreach (Control c in appID13Buttons)
                {
                    if (uapiButtons.Controls.Contains(c))
                    {
                        uapiButtons.Controls.Remove(c);
                    }
                }
                foreach (Control c in appID00Buttons)
                {
                    if (!uapiButtons.Controls.Contains(c))
                    {
                        uapiButtons.Controls.Add(c, ((uapiButtonObject)c.Tag).col, ((uapiButtonObject)c.Tag).row);
                    }
                }

                if (cfgParameters.Items.Count > 0)
                {
                    cfgParameters.Items.Clear();
                }

                foreach (ListViewItem i in appID00params)
                {
                    cfgParameters.Items.Add(i);
                }
            };

            appID13.Click += (sender, e) =>
            {
                foreach (Control c in appID00Buttons)
                {
                    if (uapiButtons.Controls.Contains(c))
                    {
                        uapiButtons.Controls.Remove(c);
                    }
                }
                foreach (Control c in appID13Buttons)
                {
                    if (!uapiButtons.Controls.Contains(c))
                    {
                        uapiButtons.Controls.Add(c, ((uapiButtonObject)c.Tag).col, ((uapiButtonObject)c.Tag).row);
                    }
                }

                if (cfgParameters.Items.Count > 0)
                {
                    cfgParameters.Items.Clear();
                }

                foreach (ListViewItem i in appID13params)
                {
                    cfgParameters.Items.Add(i);
                }
            };

            // Add the buttons to the button panel (defaulting to AppID 0x13 selected)
            uapiButtons.Controls.Add(appID00, 0, 0);
            uapiButtons.Controls.Add(appID13, 1, 0);
            foreach (Control c in appID13Buttons)
            {
                uapiButtons.Controls.Add(c, ((uapiButtonObject)c.Tag).col, ((uapiButtonObject)c.Tag).row);
            }

            // Add the parameters to the configuration parameter listview (defaulting to AppID 0x13 selected)
            foreach (ListViewItem i in appID13params)
            {
                cfgParameters.Items.Add(i);
            }

            panel.Controls.Add(uapiButtons, 0, 0);
            panel.Controls.Add(uapiMessages, 1, 0);
            panel.SetRowSpan(uapiMessages, 2);
            panel.Controls.Add(cfgParameters, 0, 1);
            panel.Controls.Add(newMessage, 0, 2);
            panel.SetColumnSpan(newMessage, 2);

            uapiBox.Controls.Add(panel);

            // First, define the overall parameter input dialog box itself
            paramBox = new Form();
            paramBox.Name = "paramBox";
            paramBox.Width = 400;
            paramBox.Height = 300;
            paramBox.Icon = Program.mf.Icon;
            paramBox.FormBorderStyle = FormBorderStyle.FixedDialog;
            paramBox.MaximizeBox = false;
            paramBox.MinimizeBox = false;

            // A container for the action button block, location button block, value textbos, and standard window buttons
            paramPanel = new TableLayoutPanel()
            {
                CellBorderStyle = TableLayoutPanelCellBorderStyle.InsetDouble,
                ColumnCount = 1,
                Dock = System.Windows.Forms.DockStyle.Fill,
                Name = "paramControls",
                RowCount = 4
            };

            paramPanel.ColumnStyles.Add(new System.Windows.Forms.ColumnStyle(System.Windows.Forms.SizeType.AutoSize));
            paramPanel.RowStyles.Add(new System.Windows.Forms.RowStyle(System.Windows.Forms.SizeType.Percent, 28.57F));
            paramPanel.RowStyles.Add(new System.Windows.Forms.RowStyle(System.Windows.Forms.SizeType.Percent, 28.57F));
            paramPanel.RowStyles.Add(new System.Windows.Forms.RowStyle(System.Windows.Forms.SizeType.Percent, 28.57F));
            paramPanel.RowStyles.Add(new System.Windows.Forms.RowStyle(System.Windows.Forms.SizeType.Percent, 14.29F));

            // This shows a manually entered parameter value to be written (set)
            setValue = new TableLayoutPanel()
            {
                ColumnCount = 1,
                Dock = System.Windows.Forms.DockStyle.Fill,
                RowCount = 2
            };
            setValue.ColumnStyles.Add(new System.Windows.Forms.ColumnStyle(System.Windows.Forms.SizeType.AutoSize));
            setValue.RowStyles.Add(new System.Windows.Forms.RowStyle(System.Windows.Forms.SizeType.Percent, 50F));
            setValue.RowStyles.Add(new System.Windows.Forms.RowStyle(System.Windows.Forms.SizeType.Percent, 50F));

            valueLabel = new Label()
            {
                Text = "Set Value"
            };
            valueLabel.Font = new Font(valueLabel.Font, FontStyle.Bold);

            setValueBox = new TextBox()
            {
                Dock = System.Windows.Forms.DockStyle.Fill,
                Enabled = false
            };
            setValueBox.Enter += (sender, e) =>
            {
                setValueBox.Clear();
            };

            setValue.Controls.Add(valueLabel, 0, 0);
            setValue.Controls.Add(setValueBox, 0, 1);

            // Three radio buttons - for selecting the desired Action
            actions = new TableLayoutPanel()
            {
                ColumnCount = 3,
                Dock = System.Windows.Forms.DockStyle.Fill,
                Text = "Action",
                RowCount = 2
            };
            actions.ColumnStyles.Add(new System.Windows.Forms.ColumnStyle(System.Windows.Forms.SizeType.Percent, 33.33F));
            actions.ColumnStyles.Add(new System.Windows.Forms.ColumnStyle(System.Windows.Forms.SizeType.Percent, 33.33F));
            actions.ColumnStyles.Add(new System.Windows.Forms.ColumnStyle(System.Windows.Forms.SizeType.Percent, 33.33F));
            actions.RowStyles.Add(new System.Windows.Forms.RowStyle(System.Windows.Forms.SizeType.Percent, 50F));
            actions.RowStyles.Add(new System.Windows.Forms.RowStyle(System.Windows.Forms.SizeType.Percent, 50F));

            actionLabel = new Label()
            {
                Text = "Actions"
            };
            actionLabel.Font = new Font(actionLabel.Font, FontStyle.Bold);

            get = new RadioButton()
            {
                Anchor = AnchorStyles.None,
                Checked = true,
                Text = "Get Value",
            };
            get.Click += (sender, e) =>
            {
                action = 0x00;
                setValueBox.Enabled = false;
            };

            set = new RadioButton()
            {
                Anchor = AnchorStyles.None,
                Checked = false,
                Text = "Set Value"
            };
            set.Click += (sender, e) =>
            {
                action = 0x40;
                if (opCode != 0x04)
                {
                    setValueBox.Enabled = true;
                }
            };

            copy = new RadioButton()
            {
                Anchor = AnchorStyles.None,
                Checked = false,
                Text = "Copy Value"
            };
            copy.Click += (sender, e) =>
            {
                action = 0x80;
                setValueBox.Enabled = false;
            };

            actions.Controls.Add(actionLabel, 0, 0);
            actions.SetColumnSpan(actionLabel, 3);
            actions.Controls.Add(get, 0, 1);
            actions.Controls.Add(set, 1, 1);
            actions.Controls.Add(copy, 2, 1);

            // Three radio buttons - for selecting the Location
            locations = new TableLayoutPanel()
            {
                ColumnCount = 3,
                Dock = System.Windows.Forms.DockStyle.Fill,
                Text = "Location",
                RowCount = 2
            };
            locations.ColumnStyles.Add(new System.Windows.Forms.ColumnStyle(System.Windows.Forms.SizeType.Percent, 33.33F));
            locations.ColumnStyles.Add(new System.Windows.Forms.ColumnStyle(System.Windows.Forms.SizeType.Percent, 33.33F));
            locations.ColumnStyles.Add(new System.Windows.Forms.ColumnStyle(System.Windows.Forms.SizeType.Percent, 33.33F));
            locations.RowStyles.Add(new System.Windows.Forms.RowStyle(System.Windows.Forms.SizeType.Percent, 50F));
            locations.RowStyles.Add(new System.Windows.Forms.RowStyle(System.Windows.Forms.SizeType.Percent, 50F));

            locationLabel = new Label()
            {
                Text = "Locations"
            };
            locationLabel.Font = new Font(locationLabel.Font, FontStyle.Bold);

            ram = new RadioButton()
            {
                Anchor = AnchorStyles.None,
                Checked = true,
                Text = "RAM"
            };
            ram.Click += (sender, e) =>
            {
                opCode = 0x01;
            };

            nvm = new RadioButton()
            {
                Anchor = AnchorStyles.None,
                Checked = false,
                Text = "NVM"
            };
            nvm.Click += (sender, e) =>
            {
                opCode = 0x02;
            };

            def = new RadioButton()
            {
                Anchor = AnchorStyles.None,
                Checked = false,
                Text = "Default"
            };
            def.Click += (sender, e) =>
            {
                opCode = 0x04;
                if (action == 0x40)
                {
                    action = 0x00;
                    get.Checked = true;
                    set.Checked = false;
                }
            };

            locations.Controls.Add(locationLabel, 0, 0);
            locations.SetColumnSpan(locationLabel, 3);
            locations.Controls.Add(ram, 0, 1);
            locations.Controls.Add(nvm, 1, 1);
            locations.Controls.Add(def, 2, 1);

            // The usual 'Ok' and 'Cancel' window buttons
            windowButtons = new TableLayoutPanel()
            {
                ColumnCount = 2,
                Dock = System.Windows.Forms.DockStyle.Fill,
                Text = null,
                RowCount = 1
            };
            windowButtons.ColumnStyles.Add(new System.Windows.Forms.ColumnStyle(System.Windows.Forms.SizeType.Percent, 50F));
            windowButtons.ColumnStyles.Add(new System.Windows.Forms.ColumnStyle(System.Windows.Forms.SizeType.Percent, 50F));
            windowButtons.RowStyles.Add(new System.Windows.Forms.RowStyle(System.Windows.Forms.SizeType.AutoSize));

            ok = new Button()
            {
                Anchor = AnchorStyles.None,
                Text = "Ok"
            };
            ok.Click += (sender, e) =>
            {
                if (cfgParameters.SelectedItems.Count > 0)
                {
                    // TODO - is this really the best way to access the highlighted parameter in the listview??
                    ListViewItem highlighted = cfgParameters.SelectedItems[0];
                    uapiParameterObject paramData = (uapiParameterObject)highlighted.Tag;
                    // If Set action, we must first convert and append the value to be written
                    if (action == 0x40)
                    {
                        byte[] value = Utilities.HexStringToByteArray(setValueBox.Text, paramData.len);

                        if (value == null)
                        {
                            return;
                        }
                        paramData.value = value;
                        Uapi.msgPayload[0] = (byte)(value.Length | action | paramData.dType);
                        value.CopyTo(Uapi.msgPayload, 2);
                        Uapi.msgPayloadLen = (ushort)(value.Length + 2);
                    }
                    else
                    {
                        Uapi.msgPayload[0] = (byte)(action | paramData.dType);  // Length is intrinsically 0 for a Get action
                        Uapi.msgPayloadLen = 2;
                    }
                    Uapi.msgPayload[1] = paramData.pID;
                    SendMessage(paramData.aID, opCode);
                }
                // TODO - catch SelectedItems.Count error here?

                paramBox.Close();
            };

            cancel = new Button()
            {
                Anchor = AnchorStyles.None,
                Text = "Cancel"
            };
            cancel.Click += (sender, e) =>
            {
                paramBox.Close();
            };

            windowButtons.Controls.Add(ok, 0, 0);
            windowButtons.Controls.Add(cancel, 1, 0);

            paramPanel.Controls.Add(actions, 0, 0);
            paramPanel.Controls.Add(locations, 0, 1);
            paramPanel.Controls.Add(setValue, 0, 2);
            paramPanel.Controls.Add(windowButtons, 0, 3);

            paramBox.Controls.Add(paramPanel);

            msgPayload = new byte[16384];
            msgPayloadLen = 0;
            opCode = 0x01;
            action = 0x00;
            uapiSeqNum = 0;
        }

        public static void uapiDialogBox()
        {
            Utilities.SetState(Constants.hostStates.HOST_STATE_UAPI_SHOWN);

            Gatt.InitializeHID();
            if (Globals.hidState == Constants.hidStates.HID_STATE_INCOMPLETE)
            {
                Uapi.uapiMessages.AppendText("Invalid HID Report data!!\n", Color.HotPink);
            }

            uapiBox.ShowDialog();
        }

        private static void ParameterDialogBox(ListViewItem highlighted)
        {
            paramBox.Text = "Parameter " + highlighted.Text + ": " + highlighted.SubItems[1].Text;

            if (!((uapiParameterObject)highlighted.Tag).rw.HasFlag(Constants.configParameterAccess.WRITE))
            {
                setValueBox.Enabled = false;
                set.Enabled = false;
            }
            else
            {
                setValueBox.Enabled = true;
                set.Enabled = true;
            }

            paramBox.ShowDialog();
        }

        public static void SendMessage(byte appID, byte opCode)
        {
            msgAppID = appID;
            msgOpCode = opCode;
            SendMessage();
        }

        public static void SendMessage()
        {
            byte[] message;

            if (txFragCnt == 0)
            {
                if (msgPayloadLen <= 15)
                {
                    message = new byte[msgPayloadLen + 5];
                    message[0] = (byte)(msgPayloadLen + 4);
                    message[1] = (byte)Constants.uapiOptions.NONE;
                    message[2] = uapiSeqNum++;
                    message[3] = msgAppID;
                    message[4] = msgOpCode;
                    if (msgPayloadLen > 0)
                    {
                        Array.Copy(msgPayload, 0, message, 5, msgPayloadLen);
                    }
                }
                else
                {
                    txFragCnt = (byte)(msgPayloadLen / 15);
                    message = new byte[20];
                    message[0] = 19;
                    message[1] = (byte)Constants.uapiOptions.START_STOP_FRAG | (byte)Constants.uapiOptions.FRAG_CONTINUE |
                                 (byte)Constants.uapiOptions.MORE_DATA;
                    message[2] = uapiSeqNum;
                    message[3] = msgAppID;
                    message[4] = txFragCnt;
                    message[5] = msgOpCode;
                    Array.Copy(msgPayload, 0, message, 6, 14);
                    msgPayloadLen -= 14;
                    msgIndex = 14;
                }
            }
            else
            {
                --txFragCnt;
                if (txFragCnt > 0)
                {
                    message = new byte[20];
                    message[0] = 19;
                    message[1] = (byte)Constants.uapiOptions.FRAG_CONTINUE | (byte)Constants.uapiOptions.MORE_DATA;
                }
                else
                {
                    message = new byte[msgPayloadLen + 5];
                    message[0] = (byte)(msgPayloadLen + 4);
                    message[1] = (byte)Constants.uapiOptions.START_STOP_FRAG;
                }
                message[2] = uapiSeqNum;
                message[3] = msgAppID;
                message[4] = txFragCnt;
                if (txFragCnt > 0)
                {
                    Array.Copy(msgPayload, msgIndex, message, 5, 15);
                    msgPayloadLen -= 15;
                    msgIndex += 15;
                }
                else
                {
                    Array.Copy(msgPayload, msgIndex, message, 5, msgPayloadLen);
                    ++uapiSeqNum;
                }
            }
            uapiMessages.AppendText(Utilities.ByteArrayToHexString(message) + "\n", Color.Green);

            CyGattWriteInfo info = new CyGattWriteInfo(Globals.uapiHidData.outputs[Globals.uapiHidData.outputIdx].dataHdl, message);
            CyApiErr err = Program.mf.gattClient.WriteCharacteristicWithoutResponse(info);
            if (err.IsNotOK)
            {
#if (debugTestUapi)
                // The following is debug-only nonsense
                Utilities.ForceException(99);
#endif
            }

        }
    }
}
